import logging

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import device, torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils.random_utils import philox_backend_seed_offset

from .topk import argsort

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
device_ = device

_MIN_INT8_VAL = tl.constexpr(torch.iinfo(torch.int8).min)
_MAX_INT8_VAL = tl.constexpr(torch.iinfo(torch.int8).max)
_MIN_INT16_VAL = tl.constexpr(torch.iinfo(torch.int16).min)
_MAX_INT16_VAL = tl.constexpr(torch.iinfo(torch.int16).max)
_MIN_INT32_VAL = tl.constexpr(torch.iinfo(torch.int32).min)
_MAX_INT32_VAL = tl.constexpr(torch.iinfo(torch.int32).max)
_MIN_INT64_VAL = tl.constexpr(torch.iinfo(torch.int64).min)
_MAX_INT64_VAL = tl.constexpr(torch.iinfo(torch.int64).max)
_MAX_UINT32_VAL = tl.constexpr((1 << 32) - 1)
_MIN_UINT32_VAL = tl.constexpr(0)
_MIN_INT24_VAL = tl.constexpr(-(2**23))
_MAX_INT24_VAL = tl.constexpr(2**23 - 1)


@triton.jit
def _get_iinfo_val(
    dtype,
    return_max,
):
    if dtype is tl.int64:
        if return_max:
            return _MAX_INT64_VAL
        else:
            return _MIN_INT64_VAL
    elif dtype is tl.int32:
        if return_max:
            return _MAX_INT32_VAL
        else:
            return _MIN_INT32_VAL
    elif dtype is tl.int16:
        if return_max:
            return _MAX_INT16_VAL
        else:
            return _MIN_INT16_VAL
    elif dtype is tl.int8:
        if return_max:
            return _MAX_INT8_VAL
        else:
            return _MIN_INT8_VAL
    elif dtype is tl.uint32:
        if return_max:
            return _MAX_UINT32_VAL
        else:
            return _MIN_UINT32_VAL
    else:
        raise ValueError("Unknown dtype")


@libentry()
@triton.jit
def bitonic_sortbykey_kernel(
    y_ptr,
    index_ptr,
    chunk_x,
    chunk_index,
    N: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    DESCENDING: tl.constexpr,
):
    cur_batch = tl.program_id(0)
    chunk_x += cur_batch * N
    chunk_index += cur_batch * N
    index_ptr += cur_batch * N
    y_ptr += cur_batch * N

    cols = tl.arange(0, BLOCK_SIZE)
    mask = cols < N

    mask_val = _get_iinfo_val(chunk_x.dtype.element_ty, return_max=not DESCENDING)

    chunk_x_val = tl.load(chunk_x + cols, mask=mask, other=mask_val)
    chunk_index_val = tl.load(chunk_index + cols, mask=mask)

    sorted_chunk_x, sorted_chunk_index = argsort(
        chunk_x_val, chunk_index_val, 0, descending=DESCENDING
    )
    tl.store(y_ptr + cols, sorted_chunk_x, mask=cols < N)
    tl.store(index_ptr + cols, sorted_chunk_index, mask=cols < N)


@triton.jit
def radix_type_convert(k):
    if tl.constexpr(k.dtype == tl.int8):
        ik = k.to(tl.int8, bitcast=True)
        mask = (ik >> 7) & 0x1
        o = tl.where(mask, ik & 0x7F, ik | 0x80)
    elif tl.constexpr(k.dtype == tl.int16):
        ik = k.to(tl.int16, bitcast=True)
        mask = (ik >> 15) & 0x1
        o = tl.where(mask, ik & 0x7FFF, ik | 0x8000)
    elif tl.constexpr(k.dtype == tl.int32):
        ik = k.to(tl.int32, bitcast=True)
        mask = (ik >> 31) & 0x1
        o = tl.where(mask, ik & 0x7FFFFFFF, ik | 0x80000000)
    elif tl.constexpr(k.dtype == tl.int64):
        ik = k.to(tl.int64, bitcast=True)
        mask = (ik >> 63) & 0x1
        o = tl.where(mask, ik & 0x7FFFFFFFFFFFFFFF, ik | 0x8000000000000000)
    else:
        o = k
    return o


@libentry()
@triton.jit
def digit_hist_kernel(
    digit_hist,
    key,
    n_elements,
    bits_per_pass,
    bins,
    passes,
    bit_mask,
    bins_segment,
    BLOCK_SIZE: tl.constexpr,
):
    bin_segid = tl.program_id(1)
    pid0 = tl.program_id(0)
    grid0 = tl.num_programs(0)

    key_offset = pid0.to(tl.int64) * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    key_mask = key_offset < n_elements
    key_data = tl.load(key + key_offset, mask=key_mask)
    ikey_data = radix_type_convert(key_data)
    bit_offset = 0
    for p in range(passes):
        key_digit = (ikey_data >> bit_offset) & bit_mask
        blk_bin_start = bin_segid * bins_segment
        for s in range(bins_segment):
            bin_id = s + blk_bin_start
            digit_mask = tl.where(key_digit == bin_id and key_mask, 1, 0)
            digit_sum = tl.sum(digit_mask)
            # +1 for exclusive
            bin_offset = p * (bins + 1) * grid0 + (bin_id + 1) * grid0 + pid0
            # reduce rather than global atomic for perf issue
            tl.store(digit_hist + bin_offset, digit_sum)
        tl.store(digit_hist + p * (bins + 1) * grid0 + pid0, 0, mask=bin_segid == 0)
        bit_offset += bits_per_pass


@libentry()
@triton.autotune(
    configs=runtime.get_tuned_config("randperm"),
    key=["n_elements"],
)
@triton.jit
def radix_sortbykey_scatter_kernel(
    key_out,
    value_out,
    key_in,
    value_in,
    digit_hist,
    d_lookback,
    n_elements,
    bit_offset,
    passes,
    p,
    num_portions,
    portion_size,
    portion_id,
    bit_mask,
    bins_segment,
    max_tiles_per_portion,
    bins: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    LOOKBACK_PARTIAL_MASK = 1 << 30
    LOOKBACK_GLOBAL_MASK = 1 << 31
    LOOKBACK_KIND_MASK = LOOKBACK_PARTIAL_MASK | LOOKBACK_GLOBAL_MASK
    LOOKBACK_VALUE_MASK = ~LOOKBACK_KIND_MASK

    pid0 = tl.program_id(0)
    portion_id_i64 = portion_id
    portion_id_i64 = portion_id_i64.to(tl.int64)
    key_offset = (
        portion_id_i64 * portion_size
        + pid0.to(tl.int64) * BLOCK_SIZE
        + tl.arange(0, BLOCK_SIZE)
    )

    key_mask = key_offset < n_elements
    value_data = tl.load(value_in + key_offset, mask=key_mask)
    key_data = tl.load(key_in + key_offset, mask=key_mask)

    ikey_data = radix_type_convert(key_data)
    key_digit = (ikey_data >> bit_offset) & bit_mask

    blk_bin_start = tl.program_id(1) * bins_segment
    last_block = tl.program_id(0) == tl.num_programs(0) - 1
    for s in range(bins_segment):
        bin_id = s + blk_bin_start
        key_digit_mask = (key_digit == bin_id) & key_mask
        key_elem_mask = tl.where(key_digit_mask, 1, 0)
        key_block_rank = tl.cumsum(key_elem_mask)
        key_block_rank = tl.where(key_digit_mask, key_block_rank - 1, 0)
        bin_of_bucket = tl.sum(key_elem_mask)
        partial_counter = bin_of_bucket | LOOKBACK_PARTIAL_MASK
        tl.store(
            d_lookback
            + ((portion_id * passes + p) * max_tiles_per_portion + pid0) * bins
            + bin_id,
            partial_counter,
            cache_modifier=".wt",
        )
        bin_offset = p * (bins + 1) + bin_id
        prefix_offsets = tl.load(
            digit_hist + bin_offset + portion_id * passes * (bins + 1)
        )
        bk = pid0 - 1
        inc_sum = bin_of_bucket
        while bk >= 0:
            rd_lbk_offset = (
                (portion_id * passes + p) * max_tiles_per_portion + bk
            ) * bins + bin_id
            partial_prefix = tl.load(d_lookback + rd_lbk_offset, volatile=True)
            while partial_prefix == 0:
                partial_prefix = tl.load(d_lookback + rd_lbk_offset, volatile=True)
            inc_sum += (partial_prefix & LOOKBACK_VALUE_MASK).to(tl.int32)
            if partial_prefix & LOOKBACK_GLOBAL_MASK:
                # break
                bk = -1
            else:
                bk -= 1
        global_counter = inc_sum | LOOKBACK_GLOBAL_MASK
        tl.store(
            d_lookback
            + ((portion_id * passes + p) * max_tiles_per_portion + pid0) * bins
            + bin_id,
            global_counter,
            cache_modifier=".wt",
        )
        inc_bucket_offset = prefix_offsets.to(tl.int64) + inc_sum.to(tl.int64)
        if last_block and portion_id < num_portions - 1:
            tl.store(
                digit_hist + bin_offset + (portion_id + 1) * passes * (bins + 1),
                inc_bucket_offset,
            )
        global_offsets = (
            inc_bucket_offset - bin_of_bucket.to(tl.int64) + key_block_rank.to(tl.int64)
        )
        tl.store(key_out + global_offsets, key_data, mask=key_digit_mask)
        tl.store(value_out + global_offsets, value_data, mask=key_digit_mask)


# for parallelization, randomly shuffle the entire block rather than adjacent equal elements as pytorch GPU backend
@libentry()
@triton.jit(do_not_specialize=["philox_seed", "philox_offset"])
def duplicate_keys_shuffle_kernel(
    value_in, n_elements, philox_seed, philox_offset, BLOCK_SIZE: tl.constexpr
):
    pid0 = tl.program_id(0)
    offset_range = tl.arange(0, BLOCK_SIZE)
    value_offset = pid0.to(tl.int64) * BLOCK_SIZE + offset_range
    value_mask = value_offset < n_elements
    value_data = tl.load(value_in + value_offset, mask=value_mask)

    philox_seed = philox_seed.to(tl.int64)
    philox_offset = philox_offset.to(tl.int64)
    c0 = (philox_offset & 0xFFFFFFFF).to(tl.uint32)
    c1 = ((philox_offset >> 32) & 0xFFFFFFFF).to(tl.uint32)
    i4 = tl.program_id(0) * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    c0 += i4
    _O = c0 * 0
    r0, _, _, _ = tl.philox(philox_seed, c0, c1, _O, _O)

    _block_size = BLOCK_SIZE
    r1 = r0 % _block_size.to(tl.uint32)
    mask_val = _get_iinfo_val(tl.uint32, True)
    r1 = tl.where(value_offset < n_elements, r1, mask_val)
    _, sorted_chunk_index = argsort(r1, offset_range, 0, descending=False)
    store_offset = pid0.to(tl.int64) * BLOCK_SIZE + sorted_chunk_index.to(tl.int64)
    tl.store(value_in + store_offset, value_data, mask=store_offset < n_elements)


def sort_by_key(key, value, valid_bits, generator=None):
    n_elements = key.numel()
    if n_elements > 2 * 1024:
        # radix method
        BLOCK_SIZE = 1024
        bits_per_pass = 4
        bits_per_segment = 3
        passes = triton.cdiv(valid_bits, bits_per_pass)
        bins = 2**bits_per_pass
        bins_per_sgement = 2**bits_per_segment
        bit_mask = bins - 1

        portion_size = 2**30  # 2 bits reserved for mask
        num_portions = triton.cdiv(n_elements, portion_size)
        max_portion_items = portion_size if num_portions > 1 else n_elements
        max_tiles_per_portion = triton.cdiv(max_portion_items, BLOCK_SIZE)

        hist_dtype = torch.int64 if num_portions > 1 else torch.int32
        grid_hist = (triton.cdiv(n_elements, BLOCK_SIZE), bins // bins_per_sgement)

        digit_hist_slice = torch.empty(
            (passes, bins + 1, grid_hist[0]), dtype=hist_dtype, device=key.device
        )

        digit_hist = torch.empty(
            (num_portions, passes, bins + 1), dtype=hist_dtype, device=key.device
        )
        d_lookback = torch.empty(
            num_portions * passes * bins * max_tiles_per_portion,
            dtype=torch.int32,
            device=key.device,
        )

        key_out_p = torch.empty_like(key)
        key_out_q = torch.empty_like(key)
        value_out_p = torch.empty_like(value)
        value_out_q = torch.empty_like(value)

        # step1
        d_lookback.zero_()
        with torch_device_fn.device(key.device):
            digit_hist_kernel[grid_hist](
                digit_hist_slice,
                key,
                n_elements,
                bits_per_pass,
                bins,
                passes,
                bit_mask,
                bins_per_sgement,
                BLOCK_SIZE,
            )

        # step2
        digit_hist_slice = torch.sum(digit_hist_slice, dim=2, keepdim=False)
        digit_hist_slice = digit_hist_slice.cumsum(dim=1)  # shape of [passes, bins + 1]
        digit_hist.copy_(digit_hist_slice)

        bit_offset = 0
        for p in range(passes):
            k_in = (key if p == 0 else key_out_p) if p % 2 == 0 else key_out_q
            v_in = (value if p == 0 else value_out_p) if p % 2 == 0 else value_out_q
            k_out = key_out_q if p % 2 == 0 else key_out_p
            v_out = value_out_q if p % 2 == 0 else value_out_p
            # step3
            for portion_id in range(num_portions):
                portion_items = min(
                    n_elements - portion_id * portion_size, portion_size
                )
                tiles_per_portion = triton.cdiv(portion_items, BLOCK_SIZE)
                grid_scatter = (tiles_per_portion, grid_hist[1])
                with torch_device_fn.device(key.device):
                    radix_sortbykey_scatter_kernel[grid_scatter](
                        k_out,
                        v_out,
                        k_in,
                        v_in,
                        digit_hist,
                        d_lookback,
                        n_elements,
                        bit_offset,
                        passes,
                        p,
                        num_portions,
                        portion_size,
                        portion_id,
                        bit_mask,
                        bins_per_sgement,
                        max_tiles_per_portion,
                        bins,
                        BLOCK_SIZE,
                    )
            bit_offset += bits_per_pass

        # last step, shuffle inner-block data
        BLOCK_SIZE_SHUFFLE = 512
        grid_shuffle = (triton.cdiv(n_elements, BLOCK_SIZE_SHUFFLE),)
        philox_seed, philox_offset = philox_backend_seed_offset(
            n_elements, generator=generator
        )
        with torch_device_fn.device(key.device):
            duplicate_keys_shuffle_kernel[grid_shuffle](
                v_out,
                n_elements,
                philox_seed,
                philox_offset,
                BLOCK_SIZE_SHUFFLE,
                num_warps=4,
            )
        return v_out
    else:
        # bitonic method
        BLOCK_SIZE = triton.next_power_of_2(n_elements)
        grid = (1,)
        k_out = torch.empty_like(key)
        v_out = torch.empty_like(value)
        with torch_device_fn.device(key.device):
            bitonic_sortbykey_kernel[grid](
                k_out, v_out, key, value, n_elements, BLOCK_SIZE, False
            )
        return v_out


def randperm(
    n,
    *,
    generator=None,
    out=None,
    dtype=torch.int64,
    layout=torch.strided,
    device=None,
    requires_grad=False,
    pin_memory=False,
):
    logger.debug("GEMS RANDPERM")
    assert dtype == torch.int16 or dtype == torch.int32 or dtype == torch.int64
    assert n <= _MAX_INT64_VAL, "n exceeds maximum int64"

    if device is None:
        device = torch.device(device_.name)
    in_range = torch.arange(n, dtype=dtype, device=device)

    u8max = 2**8
    u16max = 2**16
    u24max = 2**24
    u32max = 2**32

    if n <= u8max:
        valid_bits = 8
        key_dtype = torch.int8
        keymin = _MIN_INT8_VAL
        keymax = _MAX_INT8_VAL
    elif n <= u16max:
        valid_bits = 16
        key_dtype = torch.int16
        keymin = _MIN_INT16_VAL
        keymax = _MAX_INT16_VAL
    elif n <= u24max:
        valid_bits = 24
        key_dtype = torch.int32
        keymin = _MIN_INT24_VAL
        keymax = _MAX_INT24_VAL
    elif n <= u32max:
        valid_bits = 32
        key_dtype = torch.int32
        keymin = _MIN_INT32_VAL
        keymax = _MAX_INT32_VAL
    else:
        valid_bits = 64
        key_dtype = torch.int64
        keymin = _MIN_INT64_VAL
        keymax = _MAX_INT64_VAL

    rand_key = torch.randint(
        low=keymin, high=keymax, size=[n], dtype=key_dtype, device=device
    )
    perm_range = sort_by_key(rand_key, in_range, valid_bits, generator=generator)
    return perm_range



import logging

import torch
import triton

from flag_gems.runtime import torch_device_fn
from flag_gems.utils.random_utils import philox_backend_seed_offset

from .randn import randn_kernel

logger = logging.getLogger(
    f'flag_gems.runtime.backend._mthreads.ops.{__name__.split(".")[-1]}'
)
UNROLL = 4


def randn_like(
    x, *, dtype=None, layout=None, device=None, pin_memory=None, memory_format=None
):
    logger.debug("GEMS_MTHREADS RANDN_LIKE")
    if device is None:
        device = x.device.index
    if dtype is None:
        dtype = x.dtype
    out = torch.empty_like(x, device=device, dtype=dtype)
    N = x.numel()
    grid_fn = lambda meta: (triton.cdiv(N, meta["BLOCK"] * UNROLL),)
    # (TODO) Using Triton autotuner makes kernel parameters opaque to the caller,
    # hence we cannot obtain the per thread offset as in Pytorch.
    increment = triton.cdiv(N, UNROLL)
    philox_seed, philox_offset = philox_backend_seed_offset(increment)
    with torch_device_fn.device(x.device):
        randn_kernel[grid_fn](out, N, philox_seed, philox_offset)
    return out



import logging

import triton
import triton.language as tl

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(is_tensor=[True, True, True], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def lerp_tensor_kernel(input, end, weight):
    input32 = input.to(tl.float32)
    end32 = end.to(tl.float32)
    weight32 = weight.to(tl.float32)
    res32 = tl.where(
        tl.abs(weight32) < 0.5,
        input32 + weight32 * (end32 - input32),
        end32 - (end32 - input32) * (1 - weight32),
    )
    return res32.to(input.dtype)


@pointwise_dynamic(
    is_tensor=[True, True, False],
    dtypes=[None, None, float],
    promotion_methods=[(0, 1, "DEFAULT")],
)
@triton.jit(do_not_specialize=["weight"])
def lerp_scalar_kernel_head(input, end, weight):
    input32 = input.to(tl.float32)
    end32 = end.to(tl.float32)
    weight32 = weight.to(tl.float32)
    return (input32 + weight32 * (end32 - input32)).to(input.dtype)


@pointwise_dynamic(
    is_tensor=[True, True, False],
    dtypes=[None, None, float],
    promotion_methods=[(0, 1, "DEFAULT")],
)
@triton.jit(do_not_specialize=["weight"])
def lerp_scalar_kernel_tail(input, end, weight):
    input32 = input.to(tl.float32)
    end32 = end.to(tl.float32)
    weight32 = weight.to(tl.float32)
    return (end32 - (end32 - input32) * (1 - weight32)).to(input.dtype)


def lerp_tensor(input, end, weight):
    logger.debug("GEMS LERP TENSOR")
    out = lerp_tensor_kernel(input, end, weight)
    return out


def lerp_tensor_(input, end, weight):
    logger.debug("GEMS LERP INPLACE TENSOR")
    return lerp_tensor_kernel(input, end, weight, out0=input)


def lerp_scalar(input, end, weight):
    logger.debug("GEMS LERP TENSOR")
    if weight < 0.5:
        out = lerp_scalar_kernel_head(input, end, weight)
    else:
        out = lerp_scalar_kernel_tail(input, end, weight)
    return out


def lerp_scalar_(input, end, weight):
    logger.debug("GEMS LERP INPLACE TENSOR")
    if weight < 0.5:
        return lerp_scalar_kernel_head(input, end, weight, out0=input)
    else:
        return lerp_scalar_kernel_tail(input, end, weight, out0=input)



# core/nbx/cache.py
"""
NBX Cache Manager.

Extracts NBX containers to flat directory for fast loading.
Provisoire solution until NBX v2 format.

OPTIMIZATION:
  OLD: NBX (ZIP) -> zipfile.read() -> CPU RAM -> .to(cuda) -> VRAM = 3 copies
  NEW: NBX (ZIP) -> Extract once -> ~/.neurobrix/cache/ -> safetensors.load_file(device="cuda") = 1 copy

Parallel extraction using ThreadPoolExecutor
  - ~80% faster extraction for large models
  - 8 workers default (I/O bound, not CPU bound)
"""

import os
import json
import zipfile
import shutil
from pathlib import Path
from typing import Optional, Dict, Any, List
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor, as_completed
import threading


class NBXCache:
    """
    Manages extracted NBX cache for fast loading.

    Cache structure: ~/.neurobrix/cache/<model_name>/

    Benefits:
    - First run: Extract NBX to flat directory
    - Subsequent runs: Direct file access, no ZIP overhead
    - safetensors can load directly to GPU via mmap
    """

    DEFAULT_CACHE_DIR = Path.home() / ".neurobrix" / "cache"

    def __init__(self, cache_dir: Optional[Path] = None):
        self.cache_dir = Path(cache_dir) if cache_dir else self.DEFAULT_CACHE_DIR
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    def get_cache_path(self, nbx_path: Path) -> Path:
        """Get cache directory path for an NBX file."""
        # Use model name from path (parent directory name)
        model_name = Path(nbx_path).parent.name
        return self.cache_dir / model_name

    def is_cached(self, nbx_path: Path) -> bool:
        """Check if NBX is already extracted and valid."""
        cache_path = self.get_cache_path(nbx_path)
        manifest_path = cache_path / "manifest.json"
        meta_path = cache_path / ".cache_meta.json"

        if not manifest_path.exists():
            return False

        if not meta_path.exists():
            return False

        # Verify cache is newer than NBX file
        nbx_path = Path(nbx_path)
        if nbx_path.exists():
            nbx_mtime = nbx_path.stat().st_mtime
            cache_mtime = manifest_path.stat().st_mtime
            return cache_mtime >= nbx_mtime

        return True

    def get_cache_info(self, nbx_path: Path) -> Optional[Dict[str, Any]]:
        """Get cache metadata if exists."""
        cache_path = self.get_cache_path(nbx_path)
        meta_path = cache_path / ".cache_meta.json"

        if meta_path.exists():
            with open(meta_path) as f:
                return json.load(f)
        return None

    def extract(self, nbx_path: Path, force: bool = False) -> Path:
        """
        Extract NBX to cache directory.

        Args:
            nbx_path: Path to .nbx file
            force: Force re-extraction even if cached

        Returns:
            Path to extracted cache directory
        """
        nbx_path = Path(nbx_path)
        cache_path = self.get_cache_path(nbx_path)

        if self.is_cached(nbx_path) and not force:
            print(f"[Cache] Using cached: {cache_path}")
            return cache_path

        print(f"[Cache] Extracting {nbx_path.name} -> {cache_path}")

        # Remove old cache if exists
        if cache_path.exists():
            shutil.rmtree(cache_path)

        cache_path.mkdir(parents=True, exist_ok=True)

        # SPRINT 0 - R0.1: Parallel extraction using ThreadPoolExecutor
        # I/O bound operation benefits from parallel workers
        #
        # OPTIMIZATION (January 2026):
        # - Buffer size increased from 64KB (default) to 8MB for large files
        # - 8MB buffer is optimal for NVMe/SSD sequential writes
        # - Small files (<1MB) use smaller buffer to avoid memory waste
        LARGE_FILE_THRESHOLD = 1 * 1024 * 1024  # 1MB
        LARGE_BUFFER_SIZE = 8 * 1024 * 1024     # 8MB for safetensors
        SMALL_BUFFER_SIZE = 256 * 1024          # 256KB for small files

        with zipfile.ZipFile(nbx_path, 'r') as zf:
            members = zf.namelist()
            total = len(members)
            total_bytes = sum(zf.getinfo(m).file_size for m in members)

            # Thread-safe progress tracking
            progress_lock = threading.Lock()
            extracted_bytes = [0]  # Use list for mutable reference
            last_progress = [0]

            def extract_member(member: str) -> int:
                """Extract a single member and return its size."""
                # Ensure parent directories exist
                member_path = cache_path / member
                if member.endswith('/'):
                    member_path.mkdir(parents=True, exist_ok=True)
                    return 0

                member_path.parent.mkdir(parents=True, exist_ok=True)

                # Get file size to choose optimal buffer
                file_size = zf.getinfo(member).file_size
                buffer_size = LARGE_BUFFER_SIZE if file_size > LARGE_FILE_THRESHOLD else SMALL_BUFFER_SIZE

                # Extract file with optimized buffer
                with zf.open(member) as src:
                    with open(member_path, 'wb') as dst:
                        shutil.copyfileobj(src, dst, length=buffer_size)

                return file_size

            def update_progress(member: str, member_bytes: int):
                """Thread-safe progress update."""
                with progress_lock:
                    extracted_bytes[0] += member_bytes
                    progress = int((extracted_bytes[0] / total_bytes) * 100)
                    if progress >= last_progress[0] + 10 or 'safetensors' in member:
                        size_gb = extracted_bytes[0] / 1e9
                        print(f"[Cache] {progress}% ({size_gb:.2f}GB) - {member[:60]}...")
                        last_progress[0] = progress

            # Use 8 workers (I/O bound, not CPU bound)
            max_workers = min(8, len(members))
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                # Submit all extraction tasks
                future_to_member = {
                    executor.submit(extract_member, member): member
                    for member in members
                }

                # Process completed tasks
                for future in as_completed(future_to_member):
                    member = future_to_member[future]
                    try:
                        member_bytes = future.result()
                        update_progress(member, member_bytes)
                    except Exception as e:
                        print(f"[Cache] ERROR extracting {member}: {e}")
                        raise

        # Write cache metadata
        cache_meta = {
            "source": str(nbx_path.absolute()),
            "extracted_at": datetime.now().isoformat(),
            "file_count": total,
            "total_bytes": total_bytes,
            "total_gb": round(total_bytes / 1e9, 2),
        }
        with open(cache_path / ".cache_meta.json", 'w') as f:
            json.dump(cache_meta, f, indent=2)

        print(f"[Cache] Done: {total} files, {total_bytes/1e9:.2f}GB extracted")
        return cache_path

    def clear(self, model_name: Optional[str] = None):
        """Clear cache for a model or all models."""
        if model_name:
            cache_path = self.cache_dir / model_name
            if cache_path.exists():
                shutil.rmtree(cache_path)
                print(f"[Cache] Cleared: {model_name}")
            else:
                print(f"[Cache] Not found: {model_name}")
        else:
            if self.cache_dir.exists():
                shutil.rmtree(self.cache_dir)
                self.cache_dir.mkdir(parents=True, exist_ok=True)
                print("[Cache] Cleared all")

    def list_cached(self) -> List[Dict[str, Any]]:
        """List all cached models with metadata."""
        if not self.cache_dir.exists():
            return []

        cached = []
        for d in self.cache_dir.iterdir():
            if d.is_dir():
                meta_path = d / ".cache_meta.json"
                if meta_path.exists():
                    with open(meta_path) as f:
                        meta = json.load(f)
                    cached.append({
                        "name": d.name,
                        "path": str(d),
                        **meta
                    })
                else:
                    cached.append({
                        "name": d.name,
                        "path": str(d),
                        "extracted_at": "unknown",
                    })

        return cached

    def get_size(self, model_name: Optional[str] = None) -> int:
        """Get cache size in bytes."""
        if model_name:
            cache_path = self.cache_dir / model_name
            if not cache_path.exists():
                return 0
            return sum(f.stat().st_size for f in cache_path.rglob("*") if f.is_file())
        else:
            if not self.cache_dir.exists():
                return 0
            return sum(f.stat().st_size for f in self.cache_dir.rglob("*") if f.is_file())


# Global instance
_cache = NBXCache()


def get_cache() -> NBXCache:
    """Get global cache instance."""
    return _cache


def ensure_extracted(nbx_path: Path) -> Path:
    """Ensure NBX is extracted and return cache path.

    Handles both:
    - .nbx ZIP files: Extract to cache and return cache path
    - Already extracted directories: Validate and return directly
    """
    nbx_path = Path(nbx_path)

    # If it's a directory with manifest.json, it's already extracted
    if nbx_path.is_dir():
        manifest_path = nbx_path / "manifest.json"
        if manifest_path.exists():
            return nbx_path
        else:
            raise FileNotFoundError(
                f"Directory '{nbx_path}' is not a valid NBX cache (no manifest.json)"
            )

    # Otherwise, extract the .nbx file
    return _cache.extract(nbx_path)



# CELU - Pure Triton
# Source: Attorch
# NeuroBrix - NVIDIA Common

import torch
import triton
import triton.language as tl

@triton.jit
def _celu_kernel(input_ptr, output_ptr, n_elements, alpha, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis=0)
    offsets = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(input_ptr + offsets, mask=mask)
    
    # relu(x) + min(0, alpha * (exp(x / alpha) - 1))
    relu = tl.maximum(0, x)
    exp_term = alpha * (tl.exp(x / alpha) - 1)
    output = relu + tl.minimum(0, exp_term)
    
    tl.store(output_ptr + offsets, output, mask=mask)



# Gather - Pure Triton
# NeuroBrix - NVIDIA Common

import torch
import triton
import triton.language as tl

@triton.jit
def _gather_nd_kernel(
    x_ptr, indices_ptr, out_ptr,
    n_indices, index_dim,
    stride_x, stride_out,
    BLOCK_SIZE: tl.constexpr
):
    """General gather along specified dimension."""
    pid = tl.program_id(0)
    if pid >= n_indices: return
    idx = tl.load(indices_ptr + pid)
    for d_start in range(0, index_dim, BLOCK_SIZE):
        d_offs = d_start + tl.arange(0, BLOCK_SIZE)
        mask = d_offs < index_dim
        val = tl.load(x_ptr + idx * stride_x + d_offs, mask=mask, other=0.0)
        tl.store(out_ptr + pid * stride_out + d_offs, val, mask=mask)

@triton.jit
def _gather_kernel(
    out_ptr, x_ptr, indices_ptr,
    stride_out_row, stride_x_row, stride_idx_row,
    n_cols, BLOCK_SIZE: tl.constexpr
):
    row_idx = tl.program_id(0)
    cols = tl.arange(0, BLOCK_SIZE)
    mask = cols < n_cols
    idx_val = tl.load(indices_ptr + row_idx * stride_idx_row + cols, mask=mask, other=0)
    val = tl.load(x_ptr + row_idx * stride_x_row + idx_val, mask=mask, other=0.0)
    tl.store(out_ptr + row_idx * stride_out_row + cols, val, mask=mask)



import logging

import torch
import triton
import triton.language as tl

from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@triton.jit(do_not_specialize=["beta", "alpha"])
def addr_kernel(
    input_ptr,
    vec1_ptr,
    vec2_ptr,
    output_ptr,
    beta,
    alpha,
    M,
    N,
    stride_input_m,
    stride_input_n,
    stride_vec1,
    stride_vec2,
    stride_output_m,
    stride_output_n,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
):
    pid_m = tl.program_id(0)
    pid_n = tl.program_id(1)

    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)

    vec1_ptrs = vec1_ptr + offs_m * stride_vec1
    vec2_ptrs = vec2_ptr + offs_n * stride_vec2

    mask_m = offs_m < M
    mask_n = offs_n < N

    vec1 = tl.load(vec1_ptrs, mask=mask_m, other=0.0).to(tl.float32)
    vec2 = tl.load(vec2_ptrs, mask=mask_n, other=0.0).to(tl.float32)

    input_ptrs = (
        input_ptr + offs_m[:, None] * stride_input_m + offs_n[None, :] * stride_input_n
    )

    mask_2d = mask_m[:, None] & mask_n[None, :]
    input_val = tl.load(input_ptrs, mask=mask_2d, other=0.0).to(tl.float32)

    result = beta * input_val + alpha * (vec1[:, None] * vec2[None, :])

    output_ptrs = (
        output_ptr
        + offs_m[:, None] * stride_output_m
        + offs_n[None, :] * stride_output_n
    )
    tl.store(output_ptrs, result, mask=mask_2d)


def addr(input, vec1, vec2, *, beta=1, alpha=1):
    logger.debug("GEMS ADDR")
    if vec1.dim() != 1 or vec2.dim() != 1:
        raise ValueError("addr: expected 1-D vectors")

    M = vec1.shape[0]
    N = vec2.shape[0]
    output_shape = (M, N)

    try:
        input_broadcasted = torch.broadcast_to(input, output_shape)
    except RuntimeError:
        raise ValueError(
            f"addr: input tensor of shape {input.shape} cannot be broadcast to output shape {output_shape}"
        )
    out = torch.empty(output_shape, device=input.device, dtype=input.dtype)

    BLOCK_SIZE_M = 32
    BLOCK_SIZE_N = 32
    grid = lambda META: (
        triton.cdiv(M, BLOCK_SIZE_M),
        triton.cdiv(N, BLOCK_SIZE_N),
    )
    with torch_device_fn.device(input.device):
        addr_kernel[grid](
            input_broadcasted,
            vec1,
            vec2,
            out,
            beta,
            alpha,
            M,
            N,
            input_broadcasted.stride(0),
            input_broadcasted.stride(1),
            vec1.stride(0),
            vec2.stride(0),
            out.stride(0),
            out.stride(1),
            BLOCK_SIZE_M=BLOCK_SIZE_M,
            BLOCK_SIZE_N=BLOCK_SIZE_N,
        )
    return out



"""
Universal GELU (Fast Approx) - Triton Optimized (fp16 compatible)
"""
import torch
import triton
import triton.language as tl
from neurobrix.kernels.registry import register_kernel


@triton.jit
def _stable_tanh(x):
    """Numerically stable tanh that works with fp16."""
    # Clamp to avoid extreme values
    x_clamped = tl.minimum(tl.maximum(x, -20.0), 20.0)
    # Use the standard formula with clamped input
    exp_neg_2x = tl.exp(-2.0 * tl.abs(x_clamped))
    tanh_abs = (1.0 - exp_neg_2x) / (1.0 + exp_neg_2x)
    # Apply sign
    return tl.where(x >= 0, tanh_abs, -tanh_abs)


@triton.jit
def _gelu_kernel(x_ptr, out_ptr, n_elements, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements

    x = tl.load(x_ptr + offsets, mask=mask)
    # GELU with tanh approximation: 0.5 * x * (1 + tanh(sqrt(2/pi) * (x + 0.044715 * x^3)))
    output = 0.5 * x * (1 + _stable_tanh(0.7978845608 * (x + 0.044715 * x * x * x)))
    tl.store(out_ptr + offsets, output, mask=mask)


def _unary_impl(x: torch.Tensor) -> torch.Tensor:
    """Unary op implementation using pure Triton."""
    assert x.is_cuda, "Tensor must be on CUDA"
    x = x.contiguous()

    output = torch.empty_like(x)
    n_elements = output.numel()

    if n_elements == 0:
        return output

    BLOCK_SIZE = 1024
    grid = (triton.cdiv(n_elements, BLOCK_SIZE),)

    _gelu_kernel[grid](
        x, output,
        n_elements,
        BLOCK_SIZE=BLOCK_SIZE,
    )

    return output


@register_kernel(family="video", vendor="nvidia", tier="common", op_name="gelu")
def gelu_kernel(x: torch.Tensor, approximate: str = "tanh") -> torch.Tensor:
    """
    Gelu activation using pure Triton.
    Uses tanh approximation by default.

    Args:
        x: Input tensor
        approximate: Approximation mode (ignored, always uses tanh)

    Returns:
        Result tensor
    """
    return _unary_impl(x)



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import torch, triton, math
import triton.language as tl
from triton.runtime import driver
from ..dtypes import *

@triton.jit
def swizzle_tile_v1(pid, M, N, BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, GROUP_SIZE_M: tl.constexpr):
    grid_m     = tl.cdiv(M, BLOCK_SIZE_M)
    grid_n     = tl.cdiv(N, BLOCK_SIZE_N)
    width      = GROUP_SIZE_M * grid_n
    group_id   = pid // width
    group_size = tl.minimum(grid_m - group_id * GROUP_SIZE_M, GROUP_SIZE_M)
    pid_m      = group_id * GROUP_SIZE_M + (pid % group_size)
    pid_n      = (pid % width) // group_size
    return pid_m, pid_n

@triton.jit
def swizzle_tile_v2(pid, M, N, BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, GROUP_SIZE_M: tl.constexpr):
    grid_m     = tl.cdiv(M, BLOCK_SIZE_M)
    grid_n     = tl.cdiv(N, BLOCK_SIZE_N)
    width      = GROUP_SIZE_M * grid_m
    group_id   = pid // width
    group_size = tl.minimum(grid_n - group_id * GROUP_SIZE_M, GROUP_SIZE_M)
    pid_n      = group_id * GROUP_SIZE_M + (pid % group_size)
    pid_m      = (pid % width) // group_size
    return pid_m, pid_n

@triton.jit
def swizzle_tile_v3(pid, M, N, BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, GROUP_SIZE_M: tl.constexpr):
    pid_m  = pid % tl.cdiv(M, BLOCK_SIZE_M)
    pid_n  = pid // tl.cdiv(M, BLOCK_SIZE_M)
    grid_m = tl.cdiv(M, BLOCK_SIZE_M)
    grid_n = tl.cdiv(N, BLOCK_SIZE_N)
    return tl.swizzle2d(pid_m, pid_n, grid_m, grid_n, GROUP_SIZE_M)

@triton.jit
def swizzle_tile_persistent(tile_id, num_pid_in_group, num_pid_m, GROUP_SIZE_M: tl.constexpr): 
    group_id = tile_id // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (tile_id % group_size_m)
    pid_n = (tile_id % num_pid_in_group) // group_size_m
    return pid_m, pid_n

swizzle_tile = swizzle_tile_v1

@triton.jit
def linear_tile(pid, M, N, BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, GROUP_SIZE_M: tl.constexpr):
    pid_m = pid % tl.cdiv(M, BLOCK_SIZE_M)
    pid_n = pid // tl.cdiv(M, BLOCK_SIZE_M)
    return pid_m, pid_n

#################################################################################################################
@triton.jit
def dequantize(
    b,
    scales,
    zeros,
    q_shift,
    meta_dtype,
    unpack_mask,
    elements_per_sample: tl.constexpr,
    W_group_mode: tl.constexpr,
    zero_is_scalar: tl.constexpr,
):
    #Unpack
    if(elements_per_sample > 1):
        b = (b >> q_shift) & unpack_mask # int32 -> int32

    if(W_group_mode == 1): #Shift
        b = b.to(meta_dtype) - zeros 

    if(W_group_mode == 2):
        b = b.to(meta_dtype) * scales #Symmetric no shift (Grouped)

    if(W_group_mode == 3): #Asymmetric / Symmetric with shift(Grouped - (b - zeros) * scales)
        #b = (b - zeros) * scales
        if(zero_is_scalar):
            b = (b - zeros).to(meta_dtype) * scales
        else:
            b = (b.to(meta_dtype) - zeros) * scales

    if(W_group_mode == 4):
        b = tl.fma(b.to(meta_dtype), scales, zeros) #Asymmetric (Grouped - b*scales + zeros)

    return b

@triton.jit
def atomic_add_cas(ptr, value, Lock, mask=None, sem: tl.constexpr = "release"):    
    while tl.atomic_cas(Lock, 0, 1, sem=sem) == 1:
        pass
    tl.store(ptr, tl.load(ptr, mask=mask) + value, mask=mask)
    tl.debug_barrier()
    tl.atomic_xchg(Lock, 0)

def init_to_zero(name):
    return lambda nargs: nargs[name].zero_()

def next_power_of_2(v):
    return 2 ** int(math.ceil(math.log2(v)))

def is_divisible(dividend, divisor):
    return dividend % divisor == 0

def is_hip():
    return triton.runtime.driver.active.get_current_target().backend == "hip"

def gpu_has_more_shared_memory(ref_gpus = ["a100", "h100", "h200", "h20", "h800", "b100", "b200"]): 
    gpu_name = torch.cuda.get_device_properties(0).name.lower()
    return True in [g in gpu_name for g in ref_gpus]

def gpu_supports_float16_acc(
    ref_gpus=["5090", "5080", "5070", "5060", 
              "4090", "4080", "4070", "4060",
              "3090", "3080", "3070", "3060",
              "2080", "2070"]
):
    gpu_name = torch.cuda.get_device_properties(0).name.lower()
    return True in [g in gpu_name for g in ref_gpus]


def gpu_supports_bfloat16_atomicadd():
    #Triton tl.atomic_add doens't support bfloat16 even for Hopper and above. 
    #return torch.cuda.get_device_capability()[0] >= 9 #Hopper and above
    return False

NUM_SMS = torch.cuda.get_device_properties("cuda").multi_processor_count
def get_num_SMs(device):
    #return torch.cuda.get_device_properties(device).multi_processor_count
    return NUM_SMS #cache it to avoid driver ping - should be ok as multi-gpu systems tend to have the same devices

#Only powers of 2
def generate_autotune_lookup_v1(max_m=16384):
    return [min(2 ** int(math.ceil(math.log2(M))), max_m) if (M > 0) else 0 for M in range(max_m + 1)]

#Powers of 2 but also (power of 2 + next power of 2) / divisor, 
def generate_autotune_lookup_v2(max_m=16384, min_split=32, divisors=[2, 4], mode='next', include_vllm_config=False):
    lookup = [0] * (max_m + 1)
    autotune_vals = set()

    i = 0
    while (val := 2 ** i) <= max_m:
        autotune_vals.add(val)
        next_val = 2 ** (i + 1)
        if val >= min_split and next_val <= max_m:
            for d in divisors:
                interpolated = (val + next_val) // d
                autotune_vals.add(interpolated)
        i += 1

    if(include_vllm_config):
        autotune_vals.update([1, 2, 4, 8, 16, 24, 32, 40, 48, 56, 64, 
                             72, 80, 88, 96, 104, 112, 120, 128, 
                             136, 144, 152, 160, 168, 176, 184, 192,
                             200, 208, 216, 224, 232, 240, 248, 256, 384, 512])
    
    sorted_vals = sorted(autotune_vals)

    for m in range(max_m + 1):
        if(mode == 'next'):
            lookup[m] = min((x for x in sorted_vals if x >= m), default=None) #Next-value
        elif(mode == 'closest'):
            lookup[m] = min(sorted_vals, key=lambda x: (abs(x - m), x < m)) #Closest-Value
        else:
            raise Exception('Invalid mode.')
    return lookup

M_MAXVAL  = 4096 #1024, 4096, 16384
M_MAPPING = generate_autotune_lookup_v2(M_MAXVAL, mode='next')
def get_closest_m(M):
    return M_MAPPING[M] if M <= M_MAXVAL else M_MAXVAL
    
def get_gpu_shared_memory():
    return driver.active.utils.get_device_properties(0).get("max_shared_mem", 0)

###################################################################################
#Cached results to avoid runtime driver calls

IS_HIP = is_hip() 
NATIVE_ATOMIC = gpu_supports_bfloat16_atomicadd()


import logging

import torch
import triton

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(
    is_tensor=[True, True, True, False], promotion_methods=[(0, 1, 2, "DEFAULT")]
)
@triton.jit
def addcmul_forward(x, t1, t2, value):
    return x + value * t1 * t2


def addcmul(inp, tensor1, tensor2, *, value=1.0, out=None):
    logger.debug("GEMS ADDCMUL FORWARD")
    if out is None:
        out = torch.empty_like(inp)
    addcmul_forward(inp, tensor1, tensor2, value, out0=out)
    return out



"""
Copyright (c) 2024 by SageAttention team.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import torch
import triton
import triton.language as tl

@triton.jit
def quant_per_block_int8_kernel(Input, Output, Scale,
                                cu_seqlens_input, cu_seqlens_scale,
                                stride_ih, stride_in,
                                stride_oh, stride_on,
                                sm_scale,
                                H: tl.constexpr,
                                C: tl.constexpr, BLK: tl.constexpr):
    off_blk = tl.program_id(0)
    off_h = tl.program_id(1)
    off_b = tl.program_id(2)

    cu_seqlens_input_start = tl.load(cu_seqlens_input + off_b)
    cu_seqlens_input_end = tl.load(cu_seqlens_input + off_b + 1)

    L = cu_seqlens_input_end - cu_seqlens_input_start

    if (off_blk * BLK) >= L:
        return
    
    cu_seqlens_scale_start = tl.load(cu_seqlens_scale + off_b)

    offs_n = off_blk * BLK + tl.arange(0, BLK)
    offs_k = tl.arange(0, C)

    input_ptrs = Input + cu_seqlens_input_start * stride_in + off_h * stride_ih + offs_n[:, None] * stride_in + offs_k[None, :]
    output_ptrs = Output + cu_seqlens_input_start * stride_on + off_h * stride_oh + offs_n[:, None] * stride_on + offs_k[None, :]
    scale_ptrs = Scale + cu_seqlens_scale_start * H + off_h + off_blk * H

    x = tl.load(input_ptrs, mask=offs_n[:, None] < L)
    x = x.to(tl.float32)
    x *= sm_scale
    scale = tl.max(tl.abs(x)) / 127.
    x_int8 = x / scale
    x_int8 += 0.5 * tl.where(x_int8 >= 0, 1, -1)
    x_int8 = x_int8.to(tl.int8)
    tl.store(output_ptrs, x_int8, mask=offs_n[:, None] < L)
    tl.store(scale_ptrs, scale)

def per_block_int8(q, k, cu_seqlens_q, cu_seqlens_k, max_seqlen_q, max_seqlen_k, BLKQ=128, BLKK=64, sm_scale=None):
    q_int8 = torch.empty(q.shape, dtype=torch.int8, device=q.device)
    k_int8 = torch.empty(k.shape, dtype=torch.int8, device=k.device)

    h_qo = q.shape[1]
    h_kv = k.shape[1]
    head_dim = q.shape[-1]

    b = cu_seqlens_q.shape[0] - 1
    q_batch_len = cu_seqlens_q[1:] - cu_seqlens_q[:-1]
    k_batch_len = cu_seqlens_k[1:] - cu_seqlens_k[:-1]

    q_scale_len = (q_batch_len + BLKQ - 1) // BLKQ
    k_scale_len = (k_batch_len + BLKK - 1) // BLKK

    cu_seqlens_q_scale = torch.nn.functional.pad(torch.cumsum(q_scale_len, dim=0), (1, 0), value=0)
    cu_seqlens_k_scale = torch.nn.functional.pad(torch.cumsum(k_scale_len, dim=0), (1, 0), value=0)

    q_scale = torch.empty((cu_seqlens_q_scale[-1], h_qo), device=q.device, dtype=torch.float32)
    k_scale = torch.empty((cu_seqlens_k_scale[-1], h_kv), device=k.device, dtype=torch.float32)

    if sm_scale is None:
        sm_scale = head_dim**-0.5

    grid = ((max_seqlen_q + BLKQ - 1) // BLKQ, h_qo, b)
    quant_per_block_int8_kernel[grid](
        q, q_int8, q_scale,
        cu_seqlens_q, cu_seqlens_q_scale,
        q.stride(1), q.stride(0),
        q_int8.stride(1), q_int8.stride(0),
        sm_scale=(sm_scale * 1.44269504), H=h_qo,
        C=head_dim, BLK=BLKQ
    )

    grid = ((max_seqlen_k + BLKK - 1) // BLKK, h_kv, b)
    quant_per_block_int8_kernel[grid](
        k, k_int8, k_scale,
        cu_seqlens_k, cu_seqlens_k_scale,
        k.stride(1), k.stride(0),
        k_int8.stride(1), k_int8.stride(0),
        sm_scale=1.0, H=h_kv,
        C=head_dim, BLK=BLKK
    )

    return q_int8, q_scale, k_int8, k_scale, cu_seqlens_q_scale, cu_seqlens_k_scale



# Copyright (c) 2024, Tri Dao, Albert Gu.

import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange, repeat

try:
    from causal_conv1d import causal_conv1d_fn, causal_conv1d_update
except ImportError:
    causal_conv1d_fn, causal_conv1d_update = None, None

try:
    from causal_conv1d.causal_conv1d_varlen import causal_conv1d_varlen_states
except ImportError:
    causal_conv1d_varlen_states = None

try:
    from mamba_ssm.ops.triton.selective_state_update import selective_state_update
except ImportError:
    selective_state_update = None

from mamba_ssm.ops.triton.layernorm_gated import RMSNorm as RMSNormGated

from mamba_ssm.distributed.tensor_parallel import ColumnParallelLinear, RowParallelLinear
from mamba_ssm.distributed.distributed_utils import all_reduce, reduce_scatter

from mamba_ssm.ops.triton.ssd_combined import mamba_chunk_scan_combined
from mamba_ssm.ops.triton.ssd_combined import mamba_split_conv1d_scan_combined

from huggingface_hub import PyTorchModelHubMixin


class Mamba2(nn.Module, PyTorchModelHubMixin):
    def __init__(
        self,
        d_model,
        d_state=128,
        d_conv=4,
        conv_init=None,
        expand=2,
        headdim=64,
        d_ssm=None,  # If not None, we only apply SSM on this many dimensions, the rest uses gated MLP
        ngroups=1,
        A_init_range=(1, 16),
        D_has_hdim=False,
        rmsnorm=True,
        norm_before_gate=False,
        dt_min=0.001,
        dt_max=0.1,
        dt_init_floor=1e-4,
        dt_limit=(0.0, float("inf")),
        bias=False,
        conv_bias=True,
        # Fused kernel and sharding options
        chunk_size=256,
        use_mem_eff_path=True,
        layer_idx=None,  # Absorb kwarg for general module
        process_group=None,
        sequence_parallel=True,
        device=None,
        dtype=None,
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.d_model = d_model
        self.d_state = d_state
        self.d_conv = d_conv
        self.conv_init = conv_init
        self.expand = expand
        self.process_group = process_group
        self.sequence_parallel = sequence_parallel
        self.world_size = 1 if process_group is None else process_group.size()
        self.local_rank = 0 if process_group is None else process_group.rank()
        self.d_inner = (self.expand * self.d_model) // self.world_size
        assert self.d_inner * self.world_size == self.expand * self.d_model
        self.headdim = headdim
        self.d_ssm = self.d_inner if d_ssm is None else d_ssm // self.world_size
        assert ngroups % self.world_size == 0
        self.ngroups = ngroups // self.world_size
        assert self.d_ssm % self.headdim == 0
        self.nheads = self.d_ssm // self.headdim
        self.D_has_hdim = D_has_hdim
        self.rmsnorm = rmsnorm
        self.norm_before_gate = norm_before_gate
        self.dt_limit = dt_limit
        self.activation = "silu"
        self.chunk_size = chunk_size
        self.use_mem_eff_path = use_mem_eff_path
        self.layer_idx = layer_idx

        # Order: [z, x, B, C, dt]
        d_in_proj = 2 * self.d_inner + 2 * self.ngroups * self.d_state + self.nheads
        if self.process_group is None:
            self.in_proj = nn.Linear(self.d_model, d_in_proj, bias=bias, **factory_kwargs)
        else:
            self.in_proj = ColumnParallelLinear(self.d_model, d_in_proj * self.world_size, bias=bias,
                                                process_group=self.process_group, sequence_parallel=self.sequence_parallel,
                                                **factory_kwargs)

        conv_dim = self.d_ssm + 2 * self.ngroups * self.d_state
        self.conv1d = nn.Conv1d(
            in_channels=conv_dim,
            out_channels=conv_dim,
            bias=conv_bias,
            kernel_size=d_conv,
            groups=conv_dim,
            padding=d_conv - 1,
            **factory_kwargs,
        )
        if self.conv_init is not None:
            nn.init.uniform_(self.conv1d.weight, -self.conv_init, self.conv_init)

        self.act = nn.SiLU()

        # Initialize log dt bias
        dt = torch.exp(
            torch.rand(self.nheads, **factory_kwargs) * (math.log(dt_max) - math.log(dt_min))
            + math.log(dt_min)
        )
        dt = torch.clamp(dt, min=dt_init_floor)
        # Inverse of softplus: https://github.com/pytorch/pytorch/issues/72759
        inv_dt = dt + torch.log(-torch.expm1(-dt))
        self.dt_bias = nn.Parameter(inv_dt)
        # Just to be explicit. Without this we already don't put wd on dt_bias because of the check
        # name.endswith("bias") in param_grouping.py
        self.dt_bias._no_weight_decay = True

        assert A_init_range[0] > 0 and A_init_range[1] >= A_init_range[0]
        A = torch.empty(self.nheads, dtype=torch.float32, device=device).uniform_(*A_init_range)
        A_log = torch.log(A).to(dtype=dtype)
        self.A_log = nn.Parameter(A_log)
        self.A_log._no_weight_decay = True

        # D "skip" parameter
        self.D = nn.Parameter(torch.ones(self.d_ssm if self.D_has_hdim else self.nheads, device=device))
        self.D._no_weight_decay = True

        if self.rmsnorm:
            assert RMSNormGated is not None
            self.norm = RMSNormGated(self.d_ssm, eps=1e-5, norm_before_gate=self.norm_before_gate,
                                     group_size=self.d_ssm // ngroups, **factory_kwargs)

        if self.process_group is None:
            self.out_proj = nn.Linear(self.d_inner, self.d_model, bias=bias, **factory_kwargs)
        else:
            self.out_proj = RowParallelLinear(self.d_inner * self.world_size, self.d_model, bias=bias,
                                              process_group=self.process_group, sequence_parallel=self.sequence_parallel,
                                              **factory_kwargs)

    def forward(self, u, seqlen=None, seq_idx=None, cu_seqlens=None, inference_params=None):
        """
        u: (batch, seqlen, hidden_dim) if seqlen=None.
            If seqlen is not None, u is (batch * seqlen, hidden_dim). This is so that when we
            split u during sequence parallel, we split the batch * seqlen dimension
            (in case batch is small).
        Returns: same shape as u
        """
        seqlen_og = seqlen
        if seqlen is None:
            batch, seqlen, dim = u.shape
        else:
            batch_seqlen, dim = u.shape
            batch = batch_seqlen // seqlen

        conv_state, ssm_state = None, None
        if inference_params is not None:
            inference_batch = cu_seqlens.shape[0] - 1 if cu_seqlens is not None else batch
            conv_state, ssm_state = self._get_states_from_cache(inference_params, inference_batch)
            if inference_params.seqlen_offset > 0:
                # The states are updated inplace
                out, _, _ = self.step(u, conv_state, ssm_state)
                return out

        zxbcdt = self.in_proj(u)  # (B, L, d_in_proj) or (B * L, d_in_proj)
        if seqlen_og is not None:
            zxbcdt = rearrange(zxbcdt, "(b l) d -> b l d", l=seqlen)
        # If the model is loaded in fp16, without the .float() here, A might be -inf
        A = -torch.exp(self.A_log.float())  # (nheads) or (d_inner, d_state)
        dt_limit_kwargs = {} if self.dt_limit == (0.0, float("inf")) else dict(dt_limit=self.dt_limit)
        if self.use_mem_eff_path and inference_params is None:
            out = mamba_split_conv1d_scan_combined(
                zxbcdt,
                rearrange(self.conv1d.weight, "d 1 w -> d w"),
                self.conv1d.bias,
                self.dt_bias,
                A,
                D=rearrange(self.D, "(h p) -> h p", p=self.headdim) if self.D_has_hdim else self.D,
                chunk_size=self.chunk_size,
                seq_idx=seq_idx,
                activation=self.activation,
                rmsnorm_weight=self.norm.weight if self.rmsnorm else None,
                rmsnorm_eps=self.norm.eps if self.rmsnorm else 1e-6,
                outproj_weight=self.out_proj.weight,
                outproj_bias=self.out_proj.bias,
                headdim=None if self.D_has_hdim else self.headdim,
                ngroups=self.ngroups,
                norm_before_gate=self.norm_before_gate,
                **dt_limit_kwargs,
            )
            if seqlen_og is not None:
                out = rearrange(out, "b l d -> (b l) d")
            if self.process_group is not None:
                reduce_fn = reduce_scatter if self.sequence_parallel else all_reduce
                out = reduce_fn(out, self.process_group)
        else:
            d_mlp = (zxbcdt.shape[-1] - 2 * self.d_ssm - 2 * self.ngroups * self.d_state - self.nheads) // 2
            z0, x0, z, xBC, dt = torch.split(
                zxbcdt,
                [d_mlp, d_mlp, self.d_ssm, self.d_ssm + 2 * self.ngroups * self.d_state, self.nheads],
                dim=-1
            )
            if conv_state is not None:
                if cu_seqlens is None:
                    # If we just take xBC[:, :, -self.d_conv :], it will error if seqlen < self.d_conv
                    # Instead F.pad will pad with zeros if seqlen < self.d_conv, and truncate otherwise.
                    xBC_t = rearrange(xBC, "b l d -> b d l")
                    conv_state.copy_(F.pad(xBC_t, (self.d_conv - xBC_t.shape[-1], 0)))  # Update state (B D W)
                else:
                    assert causal_conv1d_varlen_states is not None, "varlen inference requires causal_conv1d package"
                    assert batch == 1, "varlen inference only supports batch dimension 1"
                    conv_varlen_states = causal_conv1d_varlen_states(
                        xBC.squeeze(0), cu_seqlens, state_len=conv_state.shape[-1]
                    )
                    conv_state.copy_(conv_varlen_states)
            assert self.activation in ["silu", "swish"]
            if causal_conv1d_fn is None or self.activation not in ["silu", "swish"]:
                assert seq_idx is None, "varlen conv1d requires the causal_conv1d package"
                xBC = self.act(
                    self.conv1d(xBC.transpose(1, 2)).transpose(1, 2)[:, :-(self.d_conv - 1)]
                )  # (B, L, self.d_ssm + 2 * ngroups * d_state)
            else:
                xBC = causal_conv1d_fn(
                    xBC.transpose(1, 2),
                    rearrange(self.conv1d.weight, "d 1 w -> d w"),
                    bias=self.conv1d.bias,
                    activation=self.activation,
                    seq_idx=seq_idx,
                ).transpose(1, 2)
            x, B, C = torch.split(xBC, [self.d_ssm, self.ngroups * self.d_state, self.ngroups * self.d_state], dim=-1)
            y = mamba_chunk_scan_combined(
                rearrange(x, "b l (h p) -> b l h p", p=self.headdim),
                dt,
                A,
                rearrange(B, "b l (g n) -> b l g n", g=self.ngroups),
                rearrange(C, "b l (g n) -> b l g n", g=self.ngroups),
                chunk_size=self.chunk_size,
                D=rearrange(self.D, "(h p) -> h p", p=self.headdim) if self.D_has_hdim else self.D,
                z=rearrange(z, "b l (h p) -> b l h p", p=self.headdim) if not self.rmsnorm else None,
                dt_bias=self.dt_bias,
                dt_softplus=True,
                seq_idx=seq_idx,
                cu_seqlens=cu_seqlens,
                **dt_limit_kwargs,
                return_final_states=ssm_state is not None,
                return_varlen_states=cu_seqlens is not None and inference_params is not None,
            )
            if ssm_state is not None:
                y, last_state, *rest = y
                if cu_seqlens is None:
                    ssm_state.copy_(last_state)
                else:
                    varlen_states = rest[0]
                    ssm_state.copy_(varlen_states)
            y = rearrange(y, "b l h p -> b l (h p)")
            if self.rmsnorm:
                y = self.norm(y, z)
            if d_mlp > 0:
                y = torch.cat([F.silu(z0) * x0, y], dim=-1)
            if seqlen_og is not None:
                y = rearrange(y, "b l d -> (b l) d")
            out = self.out_proj(y)
        return out

    def step(self, hidden_states, conv_state, ssm_state):
        dtype = hidden_states.dtype
        assert hidden_states.shape[1] == 1, "Only support decoding with 1 token at a time for now"
        zxbcdt = self.in_proj(hidden_states.squeeze(1))  # (B 2D)
        d_mlp = (zxbcdt.shape[-1] - 2 * self.d_ssm - 2 * self.ngroups * self.d_state - self.nheads) // 2
        z0, x0, z, xBC, dt = torch.split(
            zxbcdt,
            [d_mlp, d_mlp, self.d_ssm, self.d_ssm + 2 * self.ngroups * self.d_state, self.nheads],
            dim=-1
        )

        # Conv step
        if causal_conv1d_update is None:
            conv_state.copy_(torch.roll(conv_state, shifts=-1, dims=-1))  # Update state (B D W)
            conv_state[:, :, -1] = xBC
            xBC = torch.sum(conv_state * rearrange(self.conv1d.weight, "d 1 w -> d w"), dim=-1)  # (B D)
            if self.conv1d.bias is not None:
                xBC = xBC + self.conv1d.bias
            xBC = self.act(xBC).to(dtype=dtype)
        else:
            xBC = causal_conv1d_update(
                xBC,
                conv_state,
                rearrange(self.conv1d.weight, "d 1 w -> d w"),
                self.conv1d.bias,
                self.activation,
            )

        x, B, C = torch.split(xBC, [self.d_ssm, self.ngroups * self.d_state, self.ngroups * self.d_state], dim=-1)
        A = -torch.exp(self.A_log.float())  # (nheads,)

        # SSM step
        if selective_state_update is None:
            assert self.ngroups == 1, "Only support ngroups=1 for this inference code path"
            # Discretize A and B
            dt = F.softplus(dt + self.dt_bias.to(dtype=dt.dtype))  # (batch, nheads)
            dA = torch.exp(dt * A)  # (batch, nheads)
            x = rearrange(x, "b (h p) -> b h p", p=self.headdim)
            dBx = torch.einsum("bh,bn,bhp->bhpn", dt, B, x)
            ssm_state.copy_(ssm_state * rearrange(dA, "b h -> b h 1 1") + dBx)
            y = torch.einsum("bhpn,bn->bhp", ssm_state.to(dtype), C)
            y = y + rearrange(self.D.to(dtype), "h -> h 1") * x
            y = rearrange(y, "b h p -> b (h p)")
            if not self.rmsnorm:
                y = y * self.act(z)  # (B D)
        else:
            A = repeat(A, "h -> h p n", p=self.headdim, n=self.d_state).to(dtype=torch.float32)
            dt = repeat(dt, "b h -> b h p", p=self.headdim)
            dt_bias = repeat(self.dt_bias, "h -> h p", p=self.headdim)
            D = repeat(self.D, "h -> h p", p=self.headdim)
            B = rearrange(B, "b (g n) -> b g n", g=self.ngroups)
            C = rearrange(C, "b (g n) -> b g n", g=self.ngroups)
            x_reshaped = rearrange(x, "b (h p) -> b h p", p=self.headdim)
            if not self.rmsnorm:
                z = rearrange(z, "b (h p) -> b h p", p=self.headdim)
            y = selective_state_update(
                ssm_state, x_reshaped, dt, A, B, C, D, z=z if not self.rmsnorm else None,
                dt_bias=dt_bias, dt_softplus=True
            )
            y = rearrange(y, "b h p -> b (h p)")
        if self.rmsnorm:
            y = self.norm(y, z)
        if d_mlp > 0:
            y = torch.cat([F.silu(z0) * x0, y], dim=-1)
        out = self.out_proj(y)
        return out.unsqueeze(1), conv_state, ssm_state

    def allocate_inference_cache(self, batch_size, max_seqlen, dtype=None, **kwargs):
        device = self.out_proj.weight.device
        conv_dtype = self.conv1d.weight.dtype if dtype is None else dtype
        conv_state = torch.zeros(
            batch_size, self.d_conv, self.conv1d.weight.shape[0], device=device, dtype=conv_dtype
        ).transpose(1, 2)
        ssm_dtype = self.in_proj.weight.dtype if dtype is None else dtype
        ssm_state = torch.zeros(
            batch_size, self.nheads, self.headdim, self.d_state, device=device, dtype=ssm_dtype
        )
        return conv_state, ssm_state

    def _get_states_from_cache(self, inference_params, batch_size, initialize_states=False):
        assert self.layer_idx is not None
        if self.layer_idx not in inference_params.key_value_memory_dict:
            batch_shape = (batch_size,)
            conv_state = torch.zeros(
                batch_size,
                self.d_conv,
                self.conv1d.weight.shape[0],
                device=self.conv1d.weight.device,
                dtype=self.conv1d.weight.dtype,
            ).transpose(1, 2)
            ssm_state = torch.zeros(
                batch_size,
                self.nheads,
                self.headdim,
                self.d_state,
                device=self.in_proj.weight.device,
                dtype=self.in_proj.weight.dtype,
            )
            inference_params.key_value_memory_dict[self.layer_idx] = (conv_state, ssm_state)
        else:
            conv_state, ssm_state = inference_params.key_value_memory_dict[self.layer_idx]
            # TODO: What if batch size changes between generation, and we reuse the same states?
            if initialize_states:
                conv_state.zero_()
                ssm_state.zero_()
        return conv_state, ssm_state



"""
ATen to Kernel Operation Mapping.

Maps ATen operation names to their corresponding kernel operation names.
Used by the runtime to resolve and execute the appropriate Triton kernel.

The actual kernel files are discovered at runtime via the kernels.resolver module
which implements the tiers cascade (e.g., hopper -> ampere -> volta -> common).
"""

from typing import Dict, Optional, Callable, Any, List

from .classification import OpExecution, ATEN_CLASSIFICATION


# ATen op name -> kernel operation name (filename without .py)
# STANDARD: Snake Case
ATEN_TO_KERNEL: Dict[str, str] = {
    # Matrix Operations
    "aten::mm": "mm",
    "aten::bmm": "bmm",
    "aten::addmm": "mm",
    "aten::linear": "mm",
    "aten::matmul": "mm", 

    # Normalization
    "aten::layer_norm": "layernorm",
    "aten::native_layer_norm": "layernorm",
    "aten::group_norm": "groupnorm",
    "aten::native_group_norm": "groupnorm",
    "aten::rms_norm": "rms_norm",
    "aten::batch_norm": "batch_norm",
    "aten::instance_norm": "instance_norm",

    # Activations
    "aten::gelu": "gelu",
    "aten::gelu_": "gelu",
    "aten::silu": "silu",
    "aten::silu_": "silu",
    "aten::relu": "relu",
    "aten::relu_": "relu",
    "aten::sigmoid": "sigmoid",
    "aten::sigmoid_": "sigmoid",
    "aten::tanh": "tanh",
    "aten::tanh_": "tanh",
    "aten::softmax": "softmax",
    "aten::_softmax": "softmax",
    "aten::log_softmax": "log_softmax",
    "aten::leaky_relu": "leaky_relu",
    "aten::elu": "elu",

    # Element-wise Binary
    "aten::add": "add",
    "aten::add_": "add",
    "aten::sub": "sub",
    "aten::sub_": "sub",
    "aten::rsub": "sub",
    "aten::mul": "mul",
    "aten::mul_": "mul",
    "aten::div": "div",
    "aten::div_": "div",
    "aten::pow": "pow",
    "aten::pow_": "pow",
    "aten::maximum": "maximum",
    "aten::minimum": "minimum",
    "aten::eq": "eq",
    "aten::ne": "ne",
    "aten::lt": "lt",
    "aten::le": "le",
    "aten::gt": "gt",
    "aten::ge": "ge",
    "aten::where": "where",
    "aten::clamp": "clamp",

    # Element-wise Unary
    "aten::abs": "abs",
    "aten::neg": "neg",
    "aten::exp": "exp",
    "aten::log": "log",
    "aten::sqrt": "sqrt",
    "aten::rsqrt": "rsqrt",
    "aten::sin": "sin",
    "aten::cos": "cos",
    "aten::erf": "erf",
    "aten::reciprocal": "reciprocal",
    "aten::ceil": "ceil",
    "aten::floor": "floor",

    # Reductions
    "aten::sum": "reduce_sum",
    "aten::mean": "reduce_mean",
    "aten::max": "reduce_max",
    "aten::min": "reduce_min",
    "aten::prod": "reduce_prod",
    "aten::var": "var_mean",
    "aten::std": "std",
    "aten::argmax": "argmax",
    "aten::argmin": "argmin",
    "aten::any": "any",
    "aten::all": "all",
    "aten::cumsum": "cumsum",

    # Access / Shape
    "aten::embedding": "embedding",
    "aten::gather": "gather",
    "aten::scatter": "scatter",
    "aten::upsample_nearest2d": "resize",
    "aten::upsample_bilinear2d": "resize",
    "aten::interpolate": "resize",
    "aten::index_select": "index_select",
    "aten::masked_fill": "masked_fill",

    # Attention
    "aten::scaled_dot_product_attention": "scaled_dot_product_attention",
    "aten::_scaled_dot_product_attention": "scaled_dot_product_attention",
    "aten::_scaled_dot_product_flash_attention_for_cpu": "scaled_dot_product_attention",
    
    # Conv
    "aten::conv2d": "conv2d",
    "aten::convolution": "conv2d",
    "aten::conv1d": "conv1d",
    "aten::conv3d": "conv3d",
    "aten::conv_depthwise2d": "conv_depthwise2d",
}


def get_kernel_op_name(aten_op: str) -> Optional[str]:
    """Get the kernel operation name for an ATen op."""
    if not aten_op.startswith("aten::"):
        aten_op = f"aten::{aten_op}"

    if aten_op in ATEN_CLASSIFICATION:
        if ATEN_CLASSIFICATION[aten_op] == OpExecution.METADATA:
            return None

    return ATEN_TO_KERNEL.get(aten_op)


def run_aten_op(aten_op: str, hw: Any, family: str, *tensors, **attrs) -> Any:
    """Resolve and execute an ATen op via its Triton kernel."""
    kernel = get_kernel(aten_op, hw, family)
    if kernel is None:
        raise ValueError(f"Cannot execute METADATA op '{aten_op}' via kernel.")
    return kernel(*tensors, **attrs)


def list_supported_ops() -> Dict[str, OpExecution]:
    return dict(ATEN_CLASSIFICATION)


def list_triton_ops() -> list:
    return [op for op, exec_type in ATEN_CLASSIFICATION.items() if exec_type == OpExecution.TRITON]


def list_metadata_ops() -> list:
    return [op for op, exec_type in ATEN_CLASSIFICATION.items() if exec_type == OpExecution.METADATA]


import logging

import triton
import triton.language as tl

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, "INT_TO_FLOAT")])
@triton.jit
def exp_func(x, inplace):
    return tl.exp(x.to(tl.float32))


def exp(A):
    logger.debug("GEMS_CAMBRICON EXP")
    return exp_func(A, False)


def exp_(A):
    logger.debug("GEMS_CAMBRICON EXP_")
    return exp_func(A, True, out0=A)


# exp.out(Tensor self, *, Tensor(a!) out) -> Tensor(a!)
def exp_out(A, out):
    logger.debug("GEMS_CAMBRICON EXP_OUT")
    return exp_func(A, True, out0=out)



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang


import torch

from fla.ops.linear_attn.utils import normalize_output
from fla.ops.simple_gla import fused_chunk_simple_gla


@torch.compiler.disable
def fused_chunk_linear_attn(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    scale: float | None = None,
    initial_state: torch.Tensor | None = None,
    output_final_state: bool = False,
    normalize: bool = True,
    cu_seqlens: torch.LongTensor | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    r"""
    Args:
        q (torch.Tensor):
            queries of shape `[B, T, H, K]`.
        k (torch.Tensor):
            keys of shape `[B, T, H, K]`.
        v (torch.Tensor):
            values of shape `[B, T, H, V]`.
        scale (Optional[float]):
            Scale factor for linear attention scores.
            If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
        initial_state (Optional[torch.Tensor]):
            Initial state of shape `[B, H, K, V]`. Default: `None`.
        output_final_state (Optional[bool]):
            Whether to output the final state of shape `[B, H, K, V]`. Default: `False`.
        normalize (bool):
            Whether to normalize the output. Default: `True`.
        cu_seqlens (torch.LongTensor):
            Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
            consistent with the FlashAttention API.

    Returns:
        o (torch.Tensor):
            Outputs of shape `[B, T, H, V]`.
        final_state (torch.Tensor):
            Final state of shape `[B, H, K, V]` if `output_final_state=True` else `None`
    """
    o, final_state = fused_chunk_simple_gla(
        q=q,
        k=k,
        v=v,
        scale=scale,
        initial_state=initial_state,
        output_final_state=output_final_state,
        cu_seqlens=cu_seqlens,
    )
    if normalize:
        o = normalize_output(q * scale, k, o)
    return o, final_state



"""
Kernels for p-norm-induced losses.
"""


import triton
import triton.language as tl

from .utils import element_wise_kernel_configs


@triton.autotune(
    configs=element_wise_kernel_configs(),
    key=['size'],
)
@triton.jit
def p_loss_forward_kernel(
    input_pointer, target_pointer, output_pointer,
    param, size, p_loss: tl.constexpr, reduction: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    ):
    """
    Measures the smooth L1, L1, squared L2, or Huber loss of the difference
    between the input and target.

    Args:
        input_pointer: Pointer to the input.
            The input must be of shape [size].
        target_pointer: Pointer to the target.
            The target must be of shape [size].
        output_pointer: Pointer to a container the error is written to.
            The container must be of shape [size] if reduction is 'none',
            and otherwise of shape [size/BLOCK_SIZE].
        param: Parameter of loss function (i.e., beta or delta for smooth L1 and Huber).
        size: Number of elements in the input and target.
        p_loss: p-norm used to compute the error.
            Options are 0 for smooth L1, 1 for L1, 2 for squared L2, and 3 for Huber loss.
        reduction: Reduction strategy for the output.
            Options are 'none' for no reduction, 'mean' for averaging the error
            across all entries, and 'sum' for summing the error across all entries.
            If a reduction method is specified, the reduced result of each
            program is written to a separate index in the output container,
            which should later be summed.
        BLOCK_SIZE: Block size.
    """
    # This program processes BLOCK_SIZE rows.
    pid = tl.program_id(axis=0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offset < size

    input = tl.load(input_pointer + offset, mask=mask).to(tl.float32)
    target = tl.load(target_pointer + offset, mask=mask).to(tl.float32)
    diff = input - target

    if p_loss == 0:
        error = tl.where(diff < param, 0.5 * diff * diff / param, tl.abs(diff) - 0.5 * param)

    elif p_loss == 1:
        error = tl.abs(diff)

    elif p_loss == 2:
        error = diff * diff

    elif p_loss == 3:
        error = tl.where(diff < param, 0.5 * diff * diff, param * (tl.abs(diff) - 0.5 * param))

    if reduction == 'none':
        tl.store(output_pointer + offset, error, mask=mask)

    elif reduction == 'mean':
        tl.store(output_pointer + pid, tl.sum(error) / size)

    elif reduction == 'sum':
        tl.store(output_pointer + pid, tl.sum(error))


@triton.autotune(
    configs=element_wise_kernel_configs(),
    key=['size'],
)
@triton.jit
def p_loss_backward_kernel(
    output_grad_pointer, input_pointer, target_pointer,
    input_grad_pointer, target_grad_pointer, param, size,
    p_loss: tl.constexpr, reduction: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    ):
    """
    Calculates the input gradient of the smooth L1 norm, L1 norm, L2 norm, or Huber loss.

    Args:
        output_grad_pointer: Pointer to the error's output gradients.
            The output gradients must be a scalar or of shape [size].
        input_pointer: Pointer to the input.
            The input must be of shape [size].
        target_pointer: Pointer to the target.
            The target must be of shape [size].
        input_grad_pointer: Pointer to a container the input's gradients are written to.
            The container must be of shape [size].
        target_grad_pointer: Pointer to a container the target's gradients are written to.
            The container must be of shape [size].
        param: Parameter of loss function (i.e., beta or delta for smooth L1 and Huber).
        size: Number of elements in the input and target.
        p_loss: p-norm used to compute the error whose gradient is calculated.
            Options are 0 for smooth L1, 1 for L1, 2 for squared L2, and 3 for Huber loss.
        reduction: Reduction strategy for the output whose gradient is calculated.
            Options are 'none' for no reduction, 'mean' for averaging the error
            across all entries, and 'sum' for summing the error across all entries.
        BLOCK_SIZE: Block size.
    """
    # This program processes BLOCK_SIZE rows.
    pid = tl.program_id(axis=0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offset < size

    output_grad_mask = None
    if reduction == 'none':
        output_grad_pointer += offset
        output_grad_mask = mask

    input = tl.load(input_pointer + offset, mask=mask).to(tl.float32)
    target = tl.load(target_pointer + offset, mask=mask).to(tl.float32)
    diff = input - target
    output_grad = tl.load(output_grad_pointer, mask=output_grad_mask).to(tl.float32)

    if p_loss == 0:
        input_grad = tl.where(diff < param, diff / param, tl.where(0 <= diff, 1, -1))

    elif p_loss == 1:
        input_grad = tl.where(0 <= diff, 1, -1)

    elif p_loss == 2:
        input_grad = 2 * diff

    elif p_loss == 3:
        input_grad = tl.where(diff < param, diff, param * tl.where(0 <= diff, 1, -1))

    if reduction == 'mean':
        input_grad /= size

    input_grad *= output_grad
    tl.store(input_grad_pointer + offset, input_grad, mask=mask)
    tl.store(target_grad_pointer + offset, -input_grad, mask=mask)



import logging

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("triu"), key=["M", "N"])
@triton.jit(do_not_specialize=["diagonal"])
def triu_kernel(
    X,
    Y,
    M,
    N,
    diagonal,
    M_BLOCK_SIZE: tl.constexpr,
    N_BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    row = pid * M_BLOCK_SIZE + tl.arange(0, M_BLOCK_SIZE)[:, None]
    m_mask = row < M
    X += row * N
    Y += row * N

    for n_offset in range(0, N, N_BLOCK_SIZE):
        cols = n_offset + tl.arange(0, N_BLOCK_SIZE)[None, :]
        n_mask = cols < N
        mask = m_mask and n_mask

        x = tl.load(X + cols, mask, other=0.0)
        y = tl.where(row + diagonal <= cols, x, 0.0)
        tl.store(Y + cols, y, mask=mask)


@libentry()
@triton.autotune(
    configs=runtime.get_tuned_config("triu_batch"),
    key=["batch", "MN", "N", "diagonal"],
)
@triton.jit(do_not_specialize=["diagonal"])
def triu_batch_kernel(
    X,
    Y,
    batch,
    MN,
    N,
    diagonal,
    BATCH_BLOCK_SIZE: tl.constexpr,
    MN_BLOCK_SIZE: tl.constexpr,
):
    batch_id = tle.program_id(0)
    mn_id = tle.program_id(1)
    batch_workers = tle.num_programs(0)

    total_batch_workloads = tl.cdiv(batch, BATCH_BLOCK_SIZE)
    batch_workloads = 1
    while batch_workloads < tl.cdiv(batch, total_batch_workloads):
        batch_workloads *= 2

    for w in range(batch_workloads):
        batch_work_id = batch_id + w * batch_workers
        row = batch_work_id * BATCH_BLOCK_SIZE + tl.arange(0, BATCH_BLOCK_SIZE)[:, None]
        batch_mask = row < batch
        NX = X + row * MN
        NY = Y + row * MN

        cols = mn_id * MN_BLOCK_SIZE + tl.arange(0, MN_BLOCK_SIZE)[None, :]
        mn_mask = cols < MN
        mask = batch_mask and mn_mask
        x = tl.load(NX + cols, mask, other=0.0)
        m = cols // N
        n = cols % N
        y = tl.where(m + diagonal <= n, x, 0.0)
        tl.store(NY + cols, y, mask=mask)


INT32_MAX = torch.iinfo(torch.int32).max


def triu(A, diagonal=0):
    logger.debug("GEMS_ASCEND TRIU")
    A = A.contiguous()
    out = torch.empty_like(A)
    assert len(A.shape) > 1, "Input tensor must have at least 2 dimensions"
    M, N = A.shape[-2:]
    with torch_device_fn.device(A.device):
        if len(A.shape) == 2:
            grid = lambda meta: (triton.cdiv(M, meta["M_BLOCK_SIZE"]),)
            triu_kernel[grid](A, out, M, N, diagonal)
        else:
            batch = int(torch.numel(A) / M / N)
            B = A.view(batch, -1)

            def grid(meta):
                axis0 = triton.cdiv(batch, meta["BATCH_BLOCK_SIZE"])
                axis1 = triton.cdiv(M * N, meta["MN_BLOCK_SIZE"])
                while axis0 * axis1 >= 65536:
                    axis0 = axis0 // 2
                return (
                    axis0,
                    axis1,
                )

            triu_batch_kernel[grid](
                B,
                out,
                batch,
                M * N,
                N,
                diagonal,
            )
            out = out.view(A.shape)
    return out



"""
Copyright (c) 2024 by SageAttention team.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import torch
from typing import Any, List, Literal, Optional, Tuple, Union

from . import _fused

def per_block_int8(
    q: torch.Tensor, 
    k: torch.Tensor, 
    km: Optional[torch.Tensor] = None,
    BLKQ: int = 128, 
    BLKK: int = 64, 
    sm_scale: Optional[float] = None, 
    tensor_layout: str ="HND"
):
    """
    Quantize the query tensor `q` and the key tensor `k` with per block quantization.

    Parameters
    ----------
    q : torch.Tensor
        The query tensor. Shape:
        - If `tensor_layout` is "HND": ``[batch_size, num_qo_heads, qo_len, head_dim]``.
        - If `tensor_layout` is "NHD": ``[batch_size, qo_len, num_qo_heads, head_dim]``.

    k : torch.Tensor
        The key tensor. Shape:
        - If `tensor_layout` is "HND": ``[batch_size, num_kv_heads, kv_len, head_dim]``.
        - If `tensor_layout` is "NHD": ``[batch_size, kv_len, num_kv_heads, head_dim]``.

    km : Optional[torch.Tensor]
        The mean tensor of `k` along the sequence length dimension. Shape: ``[batch_size, num_kv_heads, head_dim]``.
        Should be of the same dtype as `k` if provided. Default is None.
    
    sm_scale : Optional[float]
        The scale factor for the softmax operation. Default is ``head_dim**-0.5``. 
        It will be multiplied by ``1.44269504`` to work together with the triton attention kernel.

    tensor_layout : str
        The tensor layout, either "HND" or "NHD".
        Default: "HND".

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
        A tuple containing:
        - The quantized query tensor. Shape: Same as `q` but with `int8` dtype.
        - The scale tensor of the query tensor. Shape: ``[batch_size, num_qo_heads, (qo_len + BLKQ - 1) // BLKQ]`` with `float32` dtype.
        - The quantized key tensor. Shape: Same as `k` but with `int8` dtype.
        - The scale tensor of the key tensor. Shape: ``[batch_size, num_kv_heads, (kv_len + BLKK - 1) // BLKK]`` with `float32` dtype.
    
    Note
    ----
    - The tensors `q` and `k` must have the dtype ``torch.float16`` or ``torch.bfloat16``
    """

    q_int8 = torch.empty(q.shape, dtype=torch.int8, device=q.device)
    k_int8 = torch.empty(k.shape, dtype=torch.int8, device=k.device)

    if tensor_layout == "HND":
        b, h_qo, qo_len, head_dim = q.shape
        _, h_kv, kv_len, _ = k.shape

    elif tensor_layout == "NHD":
        b, qo_len, h_qo, head_dim = q.shape
        _, kv_len, h_kv, _ = k.shape

    else:
        raise ValueError(f"Unknown tensor layout: {tensor_layout}")
    
    _tensor_layout = 0 if tensor_layout == "NHD" else 1

    q_scale = torch.empty((b, h_qo, (qo_len + BLKQ - 1) // BLKQ), device=q.device, dtype=torch.float32)
    k_scale = torch.empty((b, h_kv, (kv_len + BLKK - 1) // BLKK), device=q.device, dtype=torch.float32)

    if sm_scale is None:
        sm_scale = head_dim**-0.5
    
    sm_scale *= 1.44269504

    _fused.quant_per_block_int8_cuda(q, q_int8, q_scale, sm_scale, BLKQ, _tensor_layout)
    if km is not None:
        km = km.squeeze(1) if _tensor_layout == 0 else km.squeeze(2)
        _fused.quant_per_block_int8_fuse_sub_mean_cuda(k, km, k_int8, k_scale, BLKK, _tensor_layout)
    else:
        _fused.quant_per_block_int8_cuda(k, k_int8, k_scale, BLKK, _tensor_layout)

    return q_int8, q_scale, k_int8, k_scale

def per_warp_int8(
    q: torch.Tensor, 
    k: torch.Tensor,
    km: Optional[torch.Tensor] = None,
    BLKQ: int =128,
    WARPQ: int =32,
    BLKK: int =64,
    tensor_layout: str ="HND"
):
    """
    Quantize the query tensor `q` with per warp quantization and the key tensor `k` with per block quantization.
    Warp size of quantizing `q` is 16 or 32, with a block size of 64 or 128.
    Block size of quantizing `k` is 64 or 128.

    Parameters
    ----------
    q : torch.Tensor
        The query tensor. Shape:
        - If `tensor_layout` is "HND": ``[batch_size, num_qo_heads, qo_len, head_dim]``.
        - If `tensor_layout` is "NHD": ``[batch_size, qo_len, num_qo_heads, head_dim]``.

    k : torch.Tensor
        The key tensor. Shape:
        - If `tensor_layout` is "HND": ``[batch_size, num_kv_heads, kv_len, head_dim]``.
        - If `tensor_layout` is "NHD": ``[batch_size, kv_len, num_kv_heads, head_dim]``.

    km : Optional[torch.Tensor]
        The mean tensor of `k` along the sequence length dimension. Shape: ``[batch_size, num_kv_heads, head_dim]``.
        Should be of the same dtype as `k` if provided. Default is None.
    
    tensor_layout : str
        The tensor layout, either "HND" or "NHD".
        Default: "HND".

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
        A tuple containing:
        - The quantized query tensor. Shape: Same as `q` but with `int8` dtype.
        - The scale tensor of the query tensor. Shape: ``[batch_size, num_qo_heads, (qo_len + BLKQ - 1) // BLKQ * (BLKQ // WARPQ)]`` with `float32` dtype.
        - The quantized key tensor. Shape: Same as `k` but with `int8` dtype.
        - The scale tensor of the key tensor. Shape: ``[batch_size, num_kv_heads, (kv_len + BLKK - 1) // BLKK]`` with `float32` dtype.
    
    Note
    ----
    - The tensors `q` and `k` must have the dtype ``torch.float16`` or ``torch.bfloat16``
    """

    q_int8 = torch.empty(q.shape, dtype=torch.int8, device=q.device)
    k_int8 = torch.empty(k.shape, dtype=torch.int8, device=k.device)

    if tensor_layout == "HND":
        b, h_qo, qo_len, head_dim = q.shape
        _, h_kv, kv_len, _ = k.shape

    elif tensor_layout == "NHD":
        b, qo_len, h_qo, head_dim = q.shape
        _, kv_len, h_kv, _ = k.shape

    else:
        raise ValueError(f"Unknown tensor layout: {tensor_layout}")
    
    _tensor_layout = 0 if tensor_layout == "NHD" else 1

    q_scale = torch.empty((b, h_qo, ((qo_len + BLKQ - 1) // BLKQ) * (BLKQ // WARPQ)), device=q.device, dtype=torch.float32)
    k_scale = torch.empty((b, h_kv, (kv_len + BLKK - 1) // BLKK), device=q.device, dtype=torch.float32)

    _fused.quant_per_warp_int8_cuda(q, q_int8, q_scale, BLKQ, WARPQ, _tensor_layout)

    if km is not None:
        km = km.squeeze(1) if _tensor_layout == 0 else km.squeeze(2)
        _fused.quant_per_block_int8_fuse_sub_mean_cuda(k, km, k_int8, k_scale, BLKK, _tensor_layout)
    else:
        _fused.quant_per_block_int8_cuda(k, k_int8, k_scale, BLKK, _tensor_layout)
    
    return q_int8, q_scale, k_int8, k_scale

def sub_mean(
    v: torch.Tensor, 
    tensor_layout: str ="HND"
):
    """
    Calculate the mean of the tensor `v` along the sequence length dimension and subtract it from `v`. Result is stored as fp16.

    Parameters
    ----------
    v : torch.Tensor
        The input tensor. Shape:
        - If `tensor_layout` is "HND": ``[batch_size, num_kv_heads, kv_len, head_dim]``.
        - If `tensor_layout` is "NHD": ``[batch_size, kv_len, num_kv_heads, head_dim]``.

    tensor_layout : str
        The tensor layout, either "HND" or "NHD".
        Default: "HND".

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor]
        A tuple containing:
        - The tensor `v_smoothed` with the mean subtracted and stored as fp16. Shape: Same as `v` with `float16` dtype.
        - The mean tensor of `v` along the sequence length dimension. Shape: ``[batch_size, num_kv_heads, head_dim]`` with dtype same as `v`.

    Note
    ----
    - The tensors `v` must have the dtype ``torch.float16`` or ``torch.bfloat16``
    - The returned tensor `v_smoothed` will have dtype ``torch.float16`` regardless of the input dtype.
    - The returned mean tensor will have the same dtype as the input tensor.
    """

    _tensor_layout = 0 if tensor_layout == "NHD" else 1
    vm = v.mean(dim=1 if _tensor_layout == 0 else 2)

    v_smoothed = torch.empty(v.shape, dtype=torch.float16, device=v.device)
    
    # subtract mean and store the result as fp16
    _fused.sub_mean_cuda(v, vm, v_smoothed, _tensor_layout)

    return v_smoothed, vm

def per_channel_fp8(
    v: torch.Tensor,
    tensor_layout: str ="HND",
    scale_max: float = 448.0,
    smooth_v: bool = True
):
    """
    Transpose, pad and permute the tensor `v` and quantize it to fp8 with per channel quantization.
    `v` is first transposed along the head dimension and the sequence length dimension, then padded to a multiple of 64.
    After that, the tensor is permuted along the sequence length dimension by ``[0, 1, 8, 9, 2, 3, 10, 11, 4, 5, 12, 13, 6, 7, 14, 15]``.
    The quantization is done per channel, with the scale value and smooth factor calculated per channel.

    Parameters
    ----------
    v : torch.Tensor
        The input tensor. Shape:
        - If `tensor_layout` is "HND": ``[batch_size, num_kv_heads, kv_len, head_dim]``.
        - If `tensor_layout` is "NHD": ``[batch_size, kv_len, num_kv_heads, head_dim]``.

    tensor_layout : str
        The tensor layout, either "HND" or "NHD".
        Default: "HND".

    scale_max : float
        The maximum scale value for the quantization. Default is 448.0 (upper bound of E4M3 data format).

    smooth_v : bool
        Whether to smooth the quantized tensor. Default is True.

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]
        A tuple containing:
        - The quantized tensor `v_fp8`. Shape:
            - If `tensor_layout` is "HND": ``[batch_size, num_kv_heads, head_dim, (kv_len + 63) // 64 * 64]``, with `float8_e4m3fn` dtype.
            - If `tensor_layout` is "NHD": ``[batch_size, head_dim, num_kv_heads, (kv_len + 63) // 64 * 64]``, with `float8_e4m3fn` dtype.
        - The scale tensor of `v`. Shape: ``[batch_size, num_kv_heads, head_dim]`` with `float32` dtype.
        - The mean tensor of `v` along the sequence length dimension. Shape: ``[batch_size, num_kv_heads, head_dim]`` with `float32` dtype.

    Note
    ----
    - The tensors `v` must have the dtype ``torch.float16`` or ``torch.bfloat16``
    - The returned mean tensor will be None if `smooth_v` is False. Otherwise it will have dtype ``torch.float32``.
    """

    _tensor_layout = 0 if tensor_layout == "NHD" else 1

    if tensor_layout == "HND":
        b, h_kv, kv_len, head_dim = v.shape
        padded_len = (kv_len + 63) // 64 * 64
        v_transposed_permutted = torch.empty((b, h_kv, head_dim, padded_len), dtype=v.dtype, device=v.device)

    elif tensor_layout == "NHD":
        b, kv_len, h_kv, head_dim = v.shape
        padded_len = (kv_len + 63) // 64 * 64
        v_transposed_permutted = torch.empty((b, head_dim, h_kv, padded_len), dtype=v.dtype, device=v.device)
    
    _fused.transpose_pad_permute_cuda(v, v_transposed_permutted, _tensor_layout)

    v_fp8 = torch.empty(v_transposed_permutted.shape, dtype=torch.float8_e4m3fn, device=v.device)

    v_scale = torch.empty((b, h_kv, head_dim), dtype=torch.float32, device=v.device)
    vm = torch.empty((b, h_kv, head_dim), dtype=torch.float32, device=v.device)

    if smooth_v:
        _fused.mean_scale_fuse_quant_cuda(v_transposed_permutted, v_fp8, vm, v_scale, kv_len, scale_max, _tensor_layout)
        return v_fp8, v_scale, vm
    else:
        _fused.scale_fuse_quant_cuda(v_transposed_permutted, v_fp8, v_scale, kv_len, scale_max, _tensor_layout)
        return v_fp8, v_scale, None



    



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang
# Token-parallel implementation of KDA intra chunk kernel

import torch
import triton
import triton.language as tl

from fla.ops.utils.op import exp2
from fla.utils import autotune_cache_kwargs


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({'BH': BH}, num_warps=num_warps)
        for BH in [1, 2, 4, 8]
        for num_warps in [1, 2, 4, 8]
    ],
    key=["K", "H"],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T', 'N'])
def chunk_kda_fwd_kernel_intra_token_parallel(
    q,
    k,
    g,
    beta,
    Aqk,
    Akk,
    scale,
    cu_seqlens,
    N,
    T,
    H: tl.constexpr,
    K: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    BH: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_tg, i_hg = tl.program_id(0), tl.program_id(1)

    if IS_VARLEN:
        i_n = 0
        left, right = 0, N

        # Unrolled binary search (max B=2^32)
        # We can limit iterations based on expected max batch size if needed
        # 20 iterations covers B=1M, usually enough
        for _ in range(20):
            if left < right:
                mid = (left + right) // 2
                if i_tg < tl.load(cu_seqlens + mid + 1).to(tl.int32):
                    right = mid
                else:
                    left = mid + 1
        i_n = left

        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
        i_t = i_tg - bos
    else:
        bos = (i_tg // T) * T
        i_t = i_tg % T

    if i_t >= T:
        return

    i_c = i_t // BT
    i_s = (i_t % BT) // BC
    i_tc = i_c * BT
    i_ts = i_tc + i_s * BC

    q += bos * H*K
    k += bos * H*K
    g += bos * H*K
    Aqk += bos * H*BT
    Akk += bos * H*BC
    beta += bos * H

    BK: tl.constexpr = triton.next_power_of_2(K)
    o_h = tl.arange(0, BH)
    o_k = tl.arange(0, BK)
    m_h = (i_hg * BH + o_h) < H
    m_k = o_k < K

    p_q = tl.make_block_ptr(q + i_t * H*K, (H, K), (K, 1), (i_hg * BH, 0), (BH, BK), (1, 0))
    p_k = tl.make_block_ptr(k + i_t * H*K, (H, K), (K, 1), (i_hg * BH, 0), (BH, BK), (1, 0))
    p_g = tl.make_block_ptr(g + i_t * H*K, (H, K), (K, 1), (i_hg * BH, 0), (BH, BK), (1, 0))
    p_beta = tl.make_block_ptr(beta + i_t * H, (H,), (1,), (i_hg * BH,), (BH,), (0,))
    # [BH, BK]
    b_q = tl.load(p_q, boundary_check=(0, 1)).to(tl.float32)
    b_k = tl.load(p_k, boundary_check=(0, 1)).to(tl.float32)
    b_g = tl.load(p_g, boundary_check=(0, 1)).to(tl.float32)
    b_k = b_k * tl.load(p_beta, boundary_check=(0,)).to(tl.float32)[:, None]

    for j in range(i_ts, min(i_t + 1, min(T, i_ts + BC))):
        p_kj = tl.make_block_ptr(k + j * H*K, (H, K), (K, 1), (i_hg * BH, 0), (BH, BK), (1, 0))
        p_gj = tl.make_block_ptr(g + j * H*K, (H, K), (K, 1), (i_hg * BH, 0), (BH, BK), (1, 0))
        # [BH, BK]
        b_kj = tl.load(p_kj, boundary_check=(0, 1)).to(tl.float32)
        b_gj = tl.load(p_gj, boundary_check=(0, 1)).to(tl.float32)

        b_kgj = tl.where(m_k[None, :], b_kj * exp2(b_g - b_gj), 0.0)
        # [BH]
        b_Aqk = tl.sum(b_q * b_kgj, axis=1) * scale
        b_Akk = tl.where(j < i_t, tl.sum(b_k * b_kgj, axis=1), 0.0)

        tl.store(Aqk + i_t * H*BT + (i_hg * BH + o_h) * BT + j % BT, b_Aqk.to(Aqk.dtype.element_ty), mask=m_h)
        tl.store(Akk + i_t * H*BC + (i_hg * BH + o_h) * BC + j % BC, b_Akk.to(Akk.dtype.element_ty), mask=m_h)


def chunk_kda_fwd_intra_token_parallel(
    q: torch.Tensor,
    k: torch.Tensor,
    gk: torch.Tensor,
    beta: torch.Tensor,
    Aqk: torch.Tensor,
    Akk: torch.Tensor,
    scale: float,
    cu_seqlens: torch.LongTensor | None = None,
    chunk_size: int = 64,
    sub_chunk_size: int = 16,
) -> None:
    """
    Token-parallel implementation: each token gets its own thread block.
    Supports both fixed-length and variable-length sequences.
    Reduces wasted computation on padding.

    Writes directly to Aqk and Akk tensors (in-place).

    Args:
        q: [B, T, H, K]
        k: [B, T, H, K]
        gk: [B, T, H, K] cumsum of gates
        beta: [B, T, H]
        Aqk: [B, T, H, BT] output tensor to write to
        Akk: [B, T, H, BC] output tensor for diagonal blocks (fp32)
        scale: attention scale
        chunk_size: BT (default 64)
        sub_chunk_size: BC (default 16)
    """
    B, T, H, K = q.shape
    N = len(cu_seqlens) - 1 if cu_seqlens is not None else B
    BT = chunk_size
    BC = sub_chunk_size

    def grid(meta): return (B * T, triton.cdiv(H, meta['BH']))
    chunk_kda_fwd_kernel_intra_token_parallel[grid](
        q=q,
        k=k,
        g=gk,
        beta=beta,
        Aqk=Aqk,
        Akk=Akk,
        scale=scale,
        cu_seqlens=cu_seqlens,
        N=N,
        T=T,
        H=H,
        K=K,
        BT=BT,
        BC=BC,
    )
    return Aqk, Akk



import logging

import triton
import triton.language as tl

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(is_tensor=[True, False, False], promotion_methods=[(0, "DEFAULT")])
@triton.jit
def softplus_forward(x, beta, threshold):
    x_fp = x.to(tl.float32)
    z = x_fp * beta
    soft_z = tl.where(z > threshold, z, tl.log(1 + tl.exp(z)))
    out = (soft_z / beta).to(x.dtype)
    return out


def softplus(self, beta=1.0, threshold=20.0):
    logger.debug("GEMS SOFTPLUS FORWARD")
    output = softplus_forward(self, beta, threshold)
    return output



import itertools
import logging
from typing import List, Tuple, Union

import torch
import triton

from flag_gems.utils import pointwise_dynamic
from flag_gems.utils.tensor_wrapper import StridedBuffer

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


@pointwise_dynamic(is_tensor=[True], promotion_methods=[(0, "DEFAULT")])
@triton.jit
def copy_func(x):
    return x


def stack(
    tensors: Union[Tuple[torch.Tensor, ...], List[torch.Tensor]], dim: int = 0
) -> torch.Tensor:
    logger.debug("GEMS_ASCEND STACK")

    if len(tensors) == 0:
        raise RuntimeError("stack expected a non-empty TensorList")

    inp_shapes = [list(_.shape) for _ in tensors]
    inp0_shape = inp_shapes[0]
    for i, s in enumerate(inp_shapes[1:]):
        if (dim < -tensors[i + 1].dim() - 1) or (dim > tensors[i + 1].dim()):
            raise IndexError(
                "Dimension out of range (expected to be in range of [{}, {}], but got {})".format(
                    -tensors[i + 1].dim() - 1, tensors[i + 1].dim(), dim
                )
            )
        if s != inp0_shape:
            raise RuntimeError(
                f"stack expects each tensor to be equal size, but got {inp0_shape} at entry 0 and {s} at entry {i + 1}"
            )

    if dim < 0:
        dim = dim + len(inp0_shape) + 1

    in0_shape = inp0_shape[:dim] + [1] + inp0_shape[dim:]
    out_shape = inp0_shape[:dim] + [len(tensors)] + inp0_shape[dim:]
    out0 = torch.empty(out_shape, dtype=tensors[0].dtype, device=tensors[0].device)
    out0_strides = out0.stride()
    out0_offsets = list(
        itertools.accumulate([out0_strides[dim] for _ in inp_shapes[:-1]], initial=0)
    )

    for a, out0_offset in zip(tensors, out0_offsets):
        a = a.reshape(in0_shape)
        in_view = StridedBuffer(a, in0_shape, a.stride())
        out_view = StridedBuffer(out0, in0_shape, out0.stride(), offset=out0_offset)
        copy_func.instantiate(a.ndim)(in_view, out0=out_view)

    return out0



# core/dtype/converter.py
"""
Dtype Converter - Safe Dtype Conversion

Consolidates dtype conversion logic previously duplicated across:
- core/runtime/weight_loader.py
- core/io/loader.py

Prism is the master for dtype decisions. bf16→fp16 is allowed when Prism
has verified all weight values fit within ±65504 (fp16 range).
"""

import torch
from typing import Dict

from neurobrix.core.dtype.config import get_dtype_bytes


def safe_dtype_convert(
    tensor: torch.Tensor,
    target_dtype: torch.dtype,
) -> torch.Tensor:
    """
    Convert tensor dtype with safety clamping for bf16→fp16.

    Prism validates bf16→fp16 safety upstream via _scan_bf16_fp16_safety().
    This function adds clamping as a defense-in-depth measure to prevent
    inf/NaN from overflow in case any values exceed fp16 range (±65504).

    Args:
        tensor: Source tensor
        target_dtype: Target dtype (decided by Prism)

    Returns:
        Converted tensor
    """
    if tensor.dtype == target_dtype:
        return tensor

    if not tensor.is_floating_point():
        return tensor.to(target_dtype)

    # bf16→fp16: clamp to fp16 range to prevent overflow → inf → NaN
    if tensor.dtype == torch.bfloat16 and target_dtype == torch.float16:
        return tensor.clamp(-65504.0, 65504.0).to(torch.float16)

    return tensor.to(target_dtype)


def safe_dtype_convert_dict(
    tensors: Dict[str, torch.Tensor],
    target_dtype: torch.dtype,
) -> Dict[str, torch.Tensor]:
    """
    Convert all tensors in a dict to target dtype.

    Args:
        tensors: Dict of name -> tensor
        target_dtype: Target dtype (decided by Prism)

    Returns:
        Dict with converted tensors
    """
    return {
        name: safe_dtype_convert(tensor, target_dtype)
        for name, tensor in tensors.items()
    }


def calculate_dtype_multiplier(source_dtype: str, target_dtype: str) -> float:
    """
    Calculate memory multiplier for dtype conversion.

    Used by Prism for memory calculations during allocation planning.

    Args:
        source_dtype: Source dtype string (e.g., "float16")
        target_dtype: Target dtype string (e.g., "float32")

    Returns:
        Multiplier (e.g., 2.0 for fp16->fp32)
    """
    source_bytes = get_dtype_bytes(source_dtype)
    target_bytes = get_dtype_bytes(target_dtype)
    return target_bytes / source_bytes


def resolve_safe_fallback(
    requested: str,
    supported: list,
    log_prefix: str = "[Dtype]",
) -> str:
    """
    Resolve safe fallback dtype when requested is not supported.

    Prism decides the target dtype. This is a legacy fallback for non-Prism paths.
    bf16→fp16 is handled by Prism's safety scan upstream.

    Args:
        requested: Requested dtype string
        supported: List of supported dtype strings
        log_prefix: Log message prefix

    Returns:
        Safe dtype string
    """
    if requested in supported:
        return requested

    if requested == "bfloat16":
        if "float16" in supported:
            return "float16"
        return "float32"

    if requested == "float16":
        return "float32"

    return "float32"



"""
Concat kernel - Pure Triton
"""
import torch
import triton
import triton.language as tl
from neurobrix.kernels.registry import register_kernel


@triton.jit
def _copy_kernel(
    src_ptr, dst_ptr,
    n_elements,
    BLOCK_SIZE: tl.constexpr
):
    """Copy elements from src to dst."""
    pid = tl.program_id(0)
    offsets = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements

    val = tl.load(src_ptr + offsets, mask=mask)
    tl.store(dst_ptr + offsets, val, mask=mask)


def _concat_impl(tensors: list, dim: int = 0) -> torch.Tensor:
    """Concat implementation using Triton."""
    if len(tensors) == 0:
        raise ValueError("Need at least one tensor to concatenate")

    if len(tensors) == 1:
        return tensors[0]

    # Ensure all tensors are contiguous and on same device
    device = tensors[0].device
    dtype = tensors[0].dtype
    assert all(t.is_cuda for t in tensors)

    tensors = [t.contiguous() for t in tensors]

    # Normalize dim
    ndim = tensors[0].ndim
    if dim < 0:
        dim = ndim + dim

    # Calculate output shape
    out_shape = list(tensors[0].shape)
    concat_size = sum(t.shape[dim] for t in tensors)
    out_shape[dim] = concat_size

    output = torch.empty(out_shape, dtype=dtype, device=device)

    # For dim=0 concat, we can just copy contiguous blocks
    if dim == 0:
        offset = 0
        for t in tensors:
            n = t.numel()
            if n > 0:
                BLOCK = 1024
                grid = (triton.cdiv(n, BLOCK),)

                with torch.cuda.device(device):
                    _copy_kernel[grid](
                        t.view(-1), output.view(-1)[offset:],
                        n,
                        BLOCK_SIZE=BLOCK,
                        num_warps=4,
                    )
                offset += n
        return output

    # For other dims, move dim to position 0, concat, move back
    # Permute so concat dim is first
    perm = [dim] + [i for i in range(ndim) if i != dim]
    inv_perm = [0] * ndim
    for i, p in enumerate(perm):
        inv_perm[p] = i

    tensors_t = [t.permute(perm).contiguous() for t in tensors]

    # After permute, concat along dim 0
    out_shape_t = list(tensors_t[0].shape)
    out_shape_t[0] = sum(t.shape[0] for t in tensors_t)

    output_t = torch.empty(out_shape_t, dtype=dtype, device=device)

    offset = 0
    for t in tensors_t:
        n = t.numel()
        if n > 0:
            BLOCK = 1024
            grid = (triton.cdiv(n, BLOCK),)

            with torch.cuda.device(device):
                _copy_kernel[grid](
                    t.view(-1), output_t.view(-1)[offset:],
                    n,
                    BLOCK_SIZE=BLOCK,
                    num_warps=4,
                )
            offset += n

    # Permute back
    return output_t.permute(inv_perm).contiguous()


@register_kernel(family="llm", vendor="nvidia", tier="common", op_name="concat")
def concat_kernel(tensors: list, dim: int = 0) -> torch.Tensor:
    """Concatenate tensors - pure Triton."""
    if isinstance(tensors, torch.Tensor):
        return tensors
    return _concat_impl(tensors, dim)




import importlib
import logging
import os
from typing import Any, Callable, List, Mapping, Tuple

import torch

from flag_gems.ops.gather import gather
from flag_gems.utils.code_cache import code_cache_dir
from flag_gems.utils.code_utils import IndentedBuffer, write_atomic

logger = logging.getLogger(__name__)


def get_max_rank_shape(indices: List[torch.Tensor]) -> List[int]:
    # Filter out None values (basic indexing markers)
    tensor_indices = [idx for idx in indices if idx is not None]
    if len(tensor_indices) == 0:
        return []
    max_rank = max([len(index.shape) for index in tensor_indices])
    shape = [0 for _ in range(max_rank)]
    for i in range(max_rank):
        max_num = 0
        for index in tensor_indices:
            axis = len(index.shape) - 1 - i
            if axis >= 0:
                max_num = max(max_num, index.shape[axis])  #
        shape[max_rank - 1 - i] = max_num
    return shape


def broadcast_indices(indices, target_shape):
    for i, index in enumerate(indices):
        if index is not None and tuple(index.shape) != tuple(target_shape):
            indices[i] = torch.broadcast_to(index, target_shape)


def generate_imports(code: IndentedBuffer) -> IndentedBuffer:
    code.writeline("import triton")
    code.writeline("import triton.language as tl")
    code.newline()
    code.writeline("from flag_gems.utils import libentry, libtuner")
    code.writeline("from flag_gems import runtime")
    code.writeline("from flag_gems.utils.shape_utils import volume")
    code.writeline("from flag_gems.utils import triton_lang_extension as tle")

    code.newline()
    code.newline()
    return code


def generate_index_kernel(
    inp_rank, indices_len, index_rank, kernel_name: str, code: IndentedBuffer
):
    code.writeline("@libentry()")
    code.writeline("@triton.heuristics(runtime.get_heuristic_config('index'))")
    code.writeline("@triton.jit")
    code.writeline(f"def {kernel_name}(")
    with code.indent():
        args = ["input_ptr,"]
        args += [f"indices{i}_ptr," for i in range(indices_len)]
        args += ["out_ptr,"]
        args += [f"input_shape{i}," for i in range(inp_rank)]
        for i in range(indices_len):
            args += [f"indices{i}_shape{j}," for j in range(index_rank)]
        args += [f"input_stride{i}," for i in range(inp_rank)]
        for i in range(indices_len):
            args += [f"indices{i}_stride{j}," for j in range(index_rank)]
        args += [f"out_stride{i}," for i in range(index_rank + inp_rank - indices_len)]
        args += [
            "M,",
            "N,",
            "BLOCK_SIZE0: tl.constexpr,",
            "BLOCK_SIZE1: tl.constexpr,",
        ]
        code.writelines(args)
    code.writeline("):")

    with code.indent():
        code.writeline("pid0 = tle.program_id(axis=0)")
        code.writeline("pid1 = tle.program_id(axis=1)")
        code.writeline(
            "offset0 = pid0 * BLOCK_SIZE0 + tl.arange(0, BLOCK_SIZE0)[:, None]"
        )
        if inp_rank == indices_len:
            code.writeline("offset1 = pid1 * 1 + tl.arange(0, 1)[None, :]")
        else:
            code.writeline(
                "offset1 = pid1 * BLOCK_SIZE1 + tl.arange(0, BLOCK_SIZE1)[None, :]"
            )
        code.newline()
        code.writeline("cur_idx = offset0")
        for i in range(index_rank - 1, -1, -1):
            code.writeline(f"indices_idx{i} = cur_idx % indices0_shape{i}")
            code.writeline(f"cur_idx = cur_idx // indices0_shape{i}")
        code.newline()
        code.writeline("cur_idx = offset1")
        for i in range(inp_rank - 1, indices_len - 1, -1):
            code.writeline(f"input_idx{i} = cur_idx % input_shape{i}")
            code.writeline(f"cur_idx = cur_idx // input_shape{i}")
        code.newline()
        code.writeline("mask0 = offset0 < M")
        for i in range(indices_len):
            comp = [f"indices_idx{j} * indices{i}_stride{j}" for j in range(index_rank)]
            code.writeline(
                f"cur_index{i} = tl.load(indices{i}_ptr + {' + '.join(comp)}, mask=mask0, other=0)"
            )
        code.newline()
        index_mask = [
            f"(cur_index{i} >= 0) & (cur_index{i} < input_shape{i})"
            for i in range(indices_len)
        ]
        code.writeline(f"index_mask = {' & '.join(index_mask)}")
        code.writeline("mask1 = offset1 < N")
        code.writeline("mask = index_mask & mask0 & mask1")
        code.newline()
        comp = [f"cur_index{i} * input_stride{i}" for i in range(indices_len)]
        comp += [
            f"input_idx{i} * input_stride{i}" for i in range(indices_len, inp_rank)
        ]
        code.writeline(f"input_offset = {' + '.join(comp)}")
        comp = [f"indices_idx{i} * out_stride{i}" for i in range(index_rank)]
        comp += [
            f"input_idx{indices_len + i} * out_stride{index_rank + i}"
            for i in range(inp_rank - indices_len)
        ]
        code.writeline(f"out_offset = {' + '.join(comp)}")
        code.newline()
        code.writeline("cur_value = tl.load(input_ptr + input_offset , mask = mask)")
        code.writeline("tl.store(out_ptr + out_offset, cur_value, mask=mask)")

    code.newline()
    code.newline()
    return code


def generate_index_wrapper(
    inp_rank,
    indices_len,
    index_rank,
    wrapper_name: str,
    kernel_name: str,
    code: IndentedBuffer,
):
    code.writeline(f"def {wrapper_name}(input, indices, out):")
    with code.indent():
        code.writeline("input_shape = input.shape")
        code.writeline("input_stride = input.stride()")
        for i in range(indices_len):
            code.writeline(f"indices{i}_shape = indices[{i}].shape")
            code.writeline(f"indices{i}_stride = indices[{i}].stride()")
        code.writeline("out_shape = out.shape")
        code.writeline("out_stride = out.stride()")
        code.writeline("M = indices[0].numel()")
        code.writeline(f"N = volume(input_shape[{indices_len}: ])")
        code.newline()
        code.writeline("grid = lambda meta: (")
        with code.indent():
            code.writeline("triton.cdiv(M, meta['BLOCK_SIZE0']), ")
            code.writeline("triton.cdiv(N, meta['BLOCK_SIZE1']), ")
        code.writeline(")")
        code.newline()
        code.writeline(f"{kernel_name}[grid](")
        with code.indent():
            args = ["input,"]
            args += [f"indices[{i}]," for i in range(indices_len)]
            args += ["out,"]
            args += [f"input_shape[{i}]," for i in range(inp_rank)]
            for i in range(indices_len):
                args += [f"indices{i}_shape[{j}]," for j in range(index_rank)]
            args += [f"input_stride[{i}]," for i in range(inp_rank)]
            for i in range(indices_len):
                args += [f"indices{i}_stride[{j}]," for j in range(index_rank)]
            args += [
                f"out_stride[{i}]," for i in range(index_rank + inp_rank - indices_len)
            ]
            args += ["M,", "N,"]
            code.writelines(args)
        code.writeline(")")
        code.writeline("return input")
    code.newline()
    code.newline()
    return code


def generate_code(
    inputs: Tuple[Any],
    wrapper_name: str,
    kernel_name: str,
    code: IndentedBuffer,
):
    inp_rank = inputs[0].ndim
    # Filter out None values to get actual tensor indices
    tensor_indices = [idx for idx in inputs[1] if idx is not None]
    indices_len = len(tensor_indices)
    if indices_len == 0:
        raise ValueError("At least one non-None index tensor is required")
    index_rank = tensor_indices[0].ndim
    code = generate_imports(code)
    generate_index_kernel(inp_rank, indices_len, index_rank, kernel_name, code)
    generate_index_wrapper(
        inp_rank, indices_len, index_rank, wrapper_name, kernel_name, code
    )
    return code


class IndexFunction:
    def __init__(self):
        self.pid = os.getpid()
        self.overloads: Mapping[str, Callable] = {}

    def __call__(self, *args, **kwargs):
        inp, tensor_indices, out = args
        full_args = (inp, tensor_indices)

        key = self.arg_key(*full_args)
        if key in self.overloads:
            overload = self.overloads[key]
        else:
            code = IndentedBuffer()
            code = generate_code(
                full_args,
                "_index_wrapper",
                "_index_jit_function",
                code,
            )

            file_name = f"index_{key}.py"
            file_path = code_cache_dir() / file_name
            write_atomic(file_path, code.getvalue())

            spec = importlib.util.spec_from_file_location(
                f"_gen_module_rank_{key}",
                file_path,
            )

            m = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(m)
            overload = getattr(m, "_index_wrapper")
            self.overloads[key] = overload

        return overload(*args)

    def arg_key(self, *args, **kwargs):
        inp, tensor_indices = args[0], args[1]
        inp_rank = inp.ndim
        indices_len = len(tensor_indices)
        if indices_len == 0:
            index_rank = 0
        else:
            index_rank = tensor_indices[0].ndim
        return f"inp_rank_{inp_rank}_indices_len_{indices_len}_index_rank_{index_rank}"


_index_func = IndexFunction()


def index(inp, indices):
    logger.debug("GEMS INDEX")
    indices = list(indices)

    if not indices:
        raise ValueError("at least one index must be provided")

    # Step 1: Process indices (convert bool/int8 to long, handle None)
    # Following PyTorch meta implementation
    processed_indices = []
    for i, index in enumerate(indices):
        if index is not None:
            # Check dtype
            if index.dtype in [torch.int8, torch.bool]:
                # Convert boolean/int8 mask to long indices
                nonzero = index.nonzero()
                k = len(processed_indices)
                if k + index.ndim > inp.ndim:
                    raise IndexError(
                        f"too many indices for tensor of dimension {inp.ndim}"
                    )
                # Check shape matches
                for j in range(index.ndim):
                    if index.shape[j] != inp.shape[k + j]:
                        raise IndexError(
                            f"The shape of the mask {index.shape} at index {i} "
                            f"does not match the shape of the indexed tensor {inp.shape} at index {k + j}"
                        )
                # Extract indices from nonzero
                for j in range(index.ndim):
                    processed_indices.append(nonzero.select(1, j))
            elif index.dtype in [torch.long, torch.int, torch.int32, torch.int64]:
                processed_indices.append(index)
            else:
                raise TypeError(
                    "tensors used as indices must be long, int, byte or bool tensors"
                )
        else:
            processed_indices.append(None)

    indices = processed_indices

    # Check indices count
    if len(indices) > inp.ndim:
        raise IndexError(
            f"too many indices for tensor of dimension {inp.ndim} (got {len(indices)})"
        )

    # Step 2: Broadcast indices (only tensor indices, not None)
    tensor_indices = [idx for idx in indices if idx is not None]
    if tensor_indices:
        # Broadcast all tensor indices together
        if len(tensor_indices) > 1:
            tensor_indices = list(torch.broadcast_tensors(*tensor_indices))
        # Update indices list with broadcasted tensors
        tensor_idx = 0
        for i in range(len(indices)):
            if indices[i] is not None:
                indices[i] = tensor_indices[tensor_idx]
                tensor_idx += 1

    # Step 3: Add missing None indices (pad to input.ndim)
    while len(indices) < inp.ndim:
        indices.append(None)

    # Step 4: Check if has contiguous subspace
    # (all non-None tensors are adjacent)
    state = 0
    has_contiguous_subspace = False
    for index in indices:
        if state == 0:
            if index is not None:
                state = 1
        elif state == 1:
            if index is None:
                state = 2
        else:
            if index is not None:
                break
    else:
        has_contiguous_subspace = True

    # Step 5: Transpose to front if needed
    # If not contiguous, transpose input so all non-None indices come first
    if not has_contiguous_subspace:
        dims = []
        transposed_indices = []
        # First add all non-None index positions
        for i, index in enumerate(indices):
            if index is not None:
                dims.append(i)
                transposed_indices.append(index)
        # Then add all None positions
        for i, index in enumerate(indices):
            if index is None:
                dims.append(i)
                transposed_indices.append(index)
        # Permute input
        inp = inp.permute(dims)
        indices = transposed_indices

    # Step 6: Now indices have contiguous subspace
    # Calculate output shape: before_shape + replacement_shape + after_shape
    before_shape = []
    after_shape = []
    replacement_shape = []

    for dim, index in enumerate(indices):
        if index is None:
            if replacement_shape:
                # None after tensor indices -> goes to after_shape
                after_shape.append(inp.shape[dim])
            else:
                # None before tensor indices -> goes to before_shape
                before_shape.append(inp.shape[dim])
        else:
            # First tensor index determines replacement_shape
            if not replacement_shape:
                replacement_shape = list(index.shape)

    # Step 7: Build output shape and create output tensor
    out_shape = before_shape + replacement_shape + after_shape
    out = torch.empty(out_shape, dtype=inp.dtype, device=inp.device)

    # Step 8: Handle empty tensor case
    if inp.numel() == 0:
        return out

    # Step 9: Extract only tensor indices for kernel
    tensor_indices = [idx for idx in indices if idx is not None]
    if not tensor_indices:
        # All None, just reshape
        return inp.view(*out_shape)

    # Step 10: Call kernel with tensor indices
    # Note: kernel needs to handle the fact that input was potentially permuted
    # and output shape includes None dimensions
    if inp.ndim == 1 and len(tensor_indices) == 1:
        return gather(inp, 0, tensor_indices[0])

    # For mixed indexing, we need to adjust the kernel call
    # The kernel should work with the permuted input and handle output shape correctly
    _index_func(inp, tensor_indices, out)
    return out



import logging

import triton
import triton.language as tl
from triton.language.extra.mlu.libdevice import philox as _philox

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils.random_utils import (
    philox_backend_seed_offset,
    uint_to_uniform_float,
)
from flag_gems.utils.shape_utils import volume

from ..utils import TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@triton.heuristics(runtime.get_heuristic_config("uniform"))
@triton.jit(do_not_specialize=["philox_seed", "philox_offset"])
def uniform_kernel(
    out_ptr,
    N,
    philox_seed,
    philox_offset,
    from_,
    to,
    BLOCK: tl.constexpr,
):
    UNROLL: tl.constexpr = 4  # philox generate 128 random bits at a time
    philox_seed = philox_seed.to(tl.int64)
    philox_offset = philox_offset.to(tl.int64)

    pid = tl.program_id(0)
    num_jobs = tl.num_programs(0)
    i4_start = pid * BLOCK
    block_start = pid * UNROLL * BLOCK
    step = num_jobs * BLOCK * UNROLL

    for block_offset in range(block_start, N, step):
        sl = (philox_seed & 0xFFFFFFFF).to(tl.uint32)
        sh = ((philox_seed >> 32) & 0xFFFFFFFF).to(tl.uint32)
        c0 = (philox_offset & 0xFFFFFFFF).to(tl.uint32)
        c1 = ((philox_offset >> 32) & 0xFFFFFFFF).to(tl.uint32)
        r = _philox(BLOCK, sl, sh, c0 + i4_start, c1, 0, 0, 10)
        r = uint_to_uniform_float(r) * (to - from_) + from_
        r = tl.reshape(r, [UNROLL * BLOCK], can_reorder=True)

        off = block_offset + tl.arange(0, UNROLL * BLOCK)
        tl.store(out_ptr + off, r, mask=off < N)
        i4_start += num_jobs * BLOCK


UNROLL = 4


def uniform_(self, from_=0.0, to=1.0, *, generator=None):
    logger.debug("GEMS_CAMBRICON UNIFORM")
    N = volume(self.shape)
    grid_fn = lambda meta: (
        min(triton.cdiv(N, meta["BLOCK"] * UNROLL), TOTAL_CORE_NUM),
    )

    increment = triton.cdiv(N, UNROLL)
    philox_seed, philox_offset = philox_backend_seed_offset(
        increment, generator=generator
    )
    with torch_device_fn.device(self.device):
        uniform_kernel[grid_fn](
            self, N, philox_seed, philox_offset, from_, to, num_warps=1, num_stages=3
        )
    return self
